//********************************************************************************************************************************
//*                                                                                                                              *
//*     TCP Server-Client Library 1.0.40.80  3delite 2013-2025                                                                  *
//*     See TCP Server-Client Library ReadMe.txt for details.                                                                    *
//*                                                                                                                              *
//* Licenses available for usage of this component:                                                                              *
//* Freeware License: 25                                                                                                        *
//* Shareware License: 50                                                                                                       *
//* Commercial License: 250                                                                                                     *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/register.html#tcpscl                                      *
//*                                                                                                                              *
//* Home page:                                                                                                                   *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/TCPServerClientLibrary.html                               *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     https://www.3delite.hu/                                                                                                  *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit TCPClientThread;

interface

{$I TCPSCLCompilerDefines.inc}

uses
    SysUtils,
    Classes,
    SyncObjs,
    TCPLibraryClasses,
    ThreadMessageQueue,
    IdGlobal,
    IdTCPClient,
    IdComponent,
    IdSSLOpenSSL,
    HashFunctions,
    StrUtils,
    DateUtils;

type
    TTCPClientParameters = record
        Parent: TObject;
        Host: String;
        Port: Cardinal;
        LoginName: String;
        PassWord: String;
        UserName: String;
        LoggingOn: Boolean;
        SSLTLSUse: Boolean;
        LoginHashingMode: TTCPSCLHashingMode;

        Parameters: TTCPClientParams;

        ActionMessageQueue: TThreadMessageQueue;
        IncomingMessageQueue: TThreadMessageQueue;
        OnCommandReceive: TClientCommandReceiveEvent;
        OnConnected: TClientConnectedEvent;
        OnDisconnected: TClientDisconnectedEvent;
        OnCustomSend: TClientCustomSendEvent;
    end;

type
    TTCPClientThread = class(TThread)
    private
        FLoginHashingMode: TTCPSCLHashingMode;
        FSSLIOHandler: TIdSSLIOHandlerSocketOpenSSL;
        FTimeOutCleanUpDateTime: TDateTime;
        FOnCommandReceive: TClientCommandReceiveEvent;
        FOnConnected: TClientConnectedEvent;
        FOnDisconnected: TClientDisconnectedEvent;
        FOnCustomSend: TClientCustomSendEvent;

        FParameters: TTCPClientParams;

        procedure IdTCPClientConnected(Sender: TObject);
        procedure IdTCPClientDisconnected(Sender: TObject);
        procedure SendChat(Chat: TMessageChat);
        procedure SendCommand(MessageCommand: TMessageCommand);
        procedure SendBinary(MessageBinary: TMessageBinary);
        procedure ProcessTransferRequest(MessageTransferRequest: TMessageTransferRequest);
        procedure ProcessTransferCancel(MessageTransferCancel: TMessageTransferCancel);
        procedure ReportProgress(Progress: Integer; TransferID: Integer; ElapsedTime, EstimatedTime: TDateTime; TransferRate: Cardinal);
        function CheckToReceive: Boolean;
        procedure DeleteTimeOutPackets;
        procedure SendTerminated;
        function DataTransferAgreed(BinaryPacketID: Integer; List: TSendList): Boolean;
        function DataTransferCancel(BinaryPacketID: Integer; List: TSendList): Boolean;
        function DoBinaryTransfer(IdTCPClient: TIdTCPClient; MessageBinary: TMessageBinary; List: TSendList): Boolean;
    public
        Parent: TObject;
        IdTCPClient: TIdTCPClient;
        {$IFDEF INDY_ENCODING}
        ByteEncoding: IIdTextEncoding;
        {$ELSE}
        ByteEncoding: TEncoding;
        {$ENDIF}
        LoginName: String;
        PassWord: String;
        UserName: String;
        LoggingOn: Boolean;
        DownloadList: TDownloadList;
        SendList: TSendList;
        SendListSynchronous: TSendList;
        CommandList: TCommandList;
        ActionMessageQueue: TThreadMessageQueue;
        IncomingMessageQueue: TThreadMessageQueue;
        ErrorIncorrectLogIn: Boolean;
        function GenerateUniqueTransferID: Integer;
        function GenerateUniqueCommandID: Integer;
        procedure LockTransferLists;
        procedure UnLockTransferLists;
        Constructor Create(TCPClientParameters: TTCPClientParameters);
        Destructor Destroy; override;
        procedure Log(const Text: String; Debug: Boolean);
    protected
        procedure Execute; override;
    end;

implementation

constructor TTCPClientThread.Create(TCPClientParameters: TTCPClientParameters);
begin
    inherited Create(False);
    FreeOnTerminate := False;
    {$IFDEF INDY_ENCODING}
        {$IFDEF INDY_ENCODING_XE5}
        ByteEncoding := IndyTextEncoding_UTF8;
        {$ELSE}
        ByteEncoding := IndyUTF8Encoding;
        {$ENDIF}
    {$ELSE}
    ByteEncoding := TUTF8Encoding.Create;
    {$ENDIF}
    LoggingOn := TCPClientParameters.LoggingOn;
    DownloadList := TDownloadList.Create;
    SendList := TSendList.Create;
    SendListSynchronous := TSendList.Create;
    CommandList := TCommandList.Create;
    Parent := TCPClientParameters.Parent;

    FParameters := TCPClientParameters.Parameters;

    IdTCPClient := TIdTCPClient.Create;
    IdTCPClient.Host := TCPClientParameters.Host;
    IdTCPClient.Port := TCPClientParameters.Port;
    IdTCPClient.OnConnected := IdTCPClientConnected;
    IdTCPClient.OnDisconnected := IdTCPClientDisconnected;
    LoginName := TCPClientParameters.LoginName;
    PassWord := TCPClientParameters.PassWord;
    UserName := TCPClientParameters.UserName;
    ActionMessageQueue := TCPClientParameters.ActionMessageQueue;
    IncomingMessageQueue := TCPClientParameters.IncomingMessageQueue;
    FOnCommandReceive := TCPClientParameters.OnCommandReceive;
    FOnConnected := TCPClientParameters.OnConnected;
    FOnDisconnected := TCPClientParameters.OnDisconnected;
    FOnCustomSend := TCPClientParameters.OnCustomSend;
    if TCPClientParameters.SSLTLSUse then begin
        FSSLIOHandler := TIdSSLIOHandlerSocketOpenSSL.Create(nil);
        with FSSLIOHandler.SSLOptions do begin
            SSLVersions := [sslvTLSv1]; //* default
            Mode := sslmClient;
        end;
        IdTCPClient.IOHandler := FSSLIOHandler;
    end;
    FLoginHashingMode := TCPClientParameters.LoginHashingMode;
end;

destructor TTCPClientThread.Destroy;
begin
    IdTCPClient.Disconnect;
    FreeAndNil(IdTCPClient);
    if Assigned(FSSLIOHandler) then begin
        FreeAndNil(FSSLIOHandler);
    end;
    FreeAndNil(DownloadList);
    FreeAndNil(CommandList);
    FreeAndNil(SendList);
    FreeAndNil(SendListSynchronous);
    {$IFDEF INDY_ENCODING}
    //*
    {$ELSE}
    FreeAndNil(ByteEncoding);
    {$ENDIF}
    inherited;
end;

procedure TTCPClientThread.Execute;
var
    Action: TObject;
begin
    try
        ErrorIncorrectLogIn := False;
        Log('Connecting to server: ' + IdTCPClient.Host + ':' + IntToStr(IdTCPClient.Port), False);
        IdTCPClient.Connect;
    except
        on E: Exception do begin
            Log('Error connecting to server: ' + E.Message, False);
            Exit;
        end;
    end;
    try
        FTimeOutCleanUpDateTime := Now;
        while (NOT Terminated)
        AND IdTCPClient.Connected
        do begin
            //* Check out messages
            if ActionMessageQueue.Count > 0 then begin
                Action := ActionMessageQueue.Pop;
                if Action is TMessageChat then begin
                    SendChat(TMessageChat(Action));
                end else if Action is TMessageCommand then begin
                    SendCommand(TMessageCommand(Action));
                end else if Action is TMessageBinary then begin
                    SendBinary(TMessageBinary(Action));
                end else if Action is TMessageTransferRequest then begin
                    ProcessTransferRequest(TMessageTransferRequest(Action));
                end else if Action is TMessageTransferCancel then begin
                    ProcessTransferCancel(TMessageTransferCancel(Action));
                end else begin
                    //* Send custom bject
                    if Assigned(FOnCustomSend) then begin
                        FOnCustomSend(Parent, IdTCPClient, ByteEncoding, Action);
                    end else begin
                        FreeAndNil(Action);
                    end;
                end;
            end;
            if NOT CheckToReceive then begin
                //Sleep(10);
            end;
            //* Message data clean up cycle
            if MilliSecondsBetween(Now, FTimeOutCleanUpDateTime) > FParameters.TransferDataPurgeInterval then begin
                DeleteTimeOutPackets;
                FTimeOutCleanUpDateTime := Now;
            end;
        end;
    except
        On E: Exception do begin
            Log(E.Message, True);
        end;
    end;
    try
        if IdTCPClient.Connected then begin
            IdTCPClient.Disconnect;
        end else begin
            IdTCPClientDisconnected(nil);
        end;
        SendTerminated;
    except
        On E: Exception do begin
            Log(E.Message, True);
        end;
    end;
    Log('Client connection thread terminated.', True);
end;

procedure TTCPClientThread.SendTerminated;
var
    MessageTerminated: TMessageTerminated;
begin
    MessageTerminated := TMessageTerminated.Create;
    MessageTerminated.Terminated := True;
    IncomingMessageQueue.Add(MessageTerminated);
end;

function TTCPClientThread.GenerateUniqueTransferID: Integer;
var
    i: Integer;
    NewID: Integer;
    Success: Boolean;
begin
    repeat
        Success := True;
        NewID := - Random(High(Integer));
        for i := 0 to DownloadList.Count - 1 do begin
            if DownloadList.List[i].ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
        for i := 0 to SendList.Count - 1 do begin
            if TTCPServerMessageID(SendList.List[i]).ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
        for i := 0 to SendListSynchronous.Count - 1 do begin
            if TTCPServerMessageID(SendListSynchronous.List[i]).ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
    until Success
    AND (NewID <> 0);
    Result := NewID;
end;

function TTCPClientThread.GenerateUniqueCommandID: Integer;
var
    i: Integer;
    NewID: Integer;
    Success: Boolean;
begin
    repeat
        Success := True;
        NewID := - Random(High(Integer));
        for i := 0 to CommandList.Count - 1 do begin
            if CommandList.List[i].ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
    until Success
    AND (NewID <> 0);
    Result := NewID;
end;

procedure TTCPClientThread.DeleteTimeOutPackets;
var
    i: Integer;
    CurrentDateTime: TDateTime;
begin
    CurrentDateTime := Now;
    SendList.Lock;
    try
        for i := SendList.Count - 1 downto 0 do begin
            if MilliSecondsBetween(CurrentDateTime, TMessageBinary(SendList.List[i]).LastTouchTime) > FParameters.TransferDataInvalidInterval then begin
                {$IFNDEF NEXTGEN}
                TMessageBinary(SendList.List[i]).Free;
                {$ENDIF}
                SendList.Delete(i);
            end;
        end;
    finally
        SendList.UnLock;
    end;
    SendListSynchronous.Lock;
    try
        for i := SendListSynchronous.Count - 1 downto 0 do begin
            if MilliSecondsBetween(CurrentDateTime, TMessageBinary(SendListSynchronous.List[i]).LastTouchTime) > FParameters.TransferDataInvalidInterval then begin
                {$IFNDEF NEXTGEN}
                TMessageBinary(SendListSynchronous.List[i]).Free;
                {$ENDIF}
                SendListSynchronous.Delete(i);
            end;
        end;
    finally
        SendListSynchronous.UnLock;
    end;
end;

function TTCPClientThread.DataTransferAgreed(BinaryPacketID: Integer; List: TSendList): Boolean;
var
    BinaryPacketIndex: Integer;
    MessageBinary: TMessageBinary;
begin
    Result := False;
    List.Lock;
    try
        BinaryPacketIndex := List.Find(BinaryPacketID);
        if BinaryPacketIndex > - 1 then begin
            MessageBinary := TMessageBinary(List.List[BinaryPacketIndex]);
            MessageBinary.Data.Seek(MessageBinary.StartPosition, soBeginning);
            MessageBinary.DoTransfer := True;
            Result := True;
        end;
    finally
        List.UnLock;
    end;
end;

function TTCPClientThread.DataTransferCancel(BinaryPacketID: Integer; List: TSendList): Boolean;
var
    BinaryPacketIndex: Integer;
    MessageBinary: TMessageBinary;
begin
    Result := False;
    List.Lock;
    try
        BinaryPacketIndex := List.Find(BinaryPacketID);
        if BinaryPacketIndex > - 1 then begin
            MessageBinary := TMessageBinary(List.List[BinaryPacketIndex]);
            IncomingMessageQueue.Add(MessageBinary);
            Log('Binary transfer canceled: ' + IntToStr(BinaryPacketID) + ' ' + MessageBinary.FileName, True);
            List.Delete(BinaryPacketIndex);
            Result := True;
        end;
    finally
        List.UnLock;
    end;
end;

function TTCPClientThread.DoBinaryTransfer(IdTCPClient: TIdTCPClient; MessageBinary: TMessageBinary; List: TSendList): Boolean;
var
    DataSizeToSend: Integer;
    EstimatedTime: TDateTime;
    ElapsedTime: TDateTime;
    TransferRate: Cardinal;
    Index: Integer;
    FirstPacket: Boolean;
begin
    Result := False;
    //* Send the next data packet
    if MessageBinary.DoTransfer then begin
        with IdTCPClient.Socket do begin
            FirstPacket := False;
            if MessageBinary.DownloadStartTime = 0 then begin
                MessageBinary.DownloadStartTime := Now;
                MessageBinary.DownloadProgressMax := MessageBinary.TotalSize - MessageBinary.StartPosition;
                MessageBinary.DownloadLastReport := Now;
                ReportProgress(0, MessageBinary.ID, 0, 0, 0);
                Log('Upload started: ' + IntToStr(MessageBinary.ID) + ' ' + MessageBinary.FileName, True);
                FirstPacket := True;
            end;
            if MessageBinary.Data.Size - MessageBinary.Data.Position > FParameters.DataTransferPacketSize then begin
                DataSizeToSend := FParameters.DataTransferPacketSize;
            end else begin
                DataSizeToSend := MessageBinary.Data.Size - MessageBinary.Data.Position;
            end;
            if FirstPacket then begin
                WriteLn(IntToStr(MessageBinary.ID) + ' 802 ' + IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size) + ' "' + MessageBinary.FileName + '"' + ' ' + MessageBinary.Arguments, ByteEncoding);
            end else begin
                WriteLn(IntToStr(MessageBinary.ID) + ' 802' {+ IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size)}, ByteEncoding);
            end;
            WriteLn(IntToStr(DataSizeToSend), ByteEncoding);
            try
                IdTCPClient.Socket.WriteBufferOpen;
                Write(MessageBinary.Data, DataSizeToSend);
                IdTCPClient.Socket.WriteBufferClose;
                IdTCPClient.Socket.WriteBufferFlush;
            except
                on E: Exception do begin
                    Log('Exception: ' + E.Message, True);
                end;
            end;
            with MessageBinary do begin
                LastTouchTime := Now;
                //* Report every 250ms
                if DateTimeDiffmSec(DownloadLastReport, Now) > 250 then begin
                    DownloadProgress := Round((Data.Position / DownloadProgressMax) * 100);
                    EstimatedTime := ((Now - DownloadStartTime) * DownloadProgressMax) / (Data.Position - MessageBinary.StartPosition);
                    //IncSecond(EstimatedTime, - 1);
                    ElapsedTime := Now - DownloadStartTime;
                    if ElapsedTime <> 0 then begin
                        TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition) / SecondSpan(ElapsedTime, 0));
                    end else begin
                        TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
                    end;
                    ReportProgress(DownloadProgress, MessageBinary.ID, ElapsedTime, EstimatedTime, TransferRate);
                    DownloadLastReport := Now;
                end;
            end;
            Result := True;
        end;
        //* Completed
        if MessageBinary.Data.Position = MessageBinary.Data.Size then begin
            MessageBinary.DownloadProgress := 100;
            ElapsedTime := Now - MessageBinary.DownloadStartTime;
            if ElapsedTime <> 0 then begin
                TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition)/ SecondSpan(ElapsedTime, 0));
            end else begin
                TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
            end;
            ReportProgress(100, MessageBinary.ID, ElapsedTime, 0, TransferRate);
            Log('Successfully sent binary: ' + IntToStr(MessageBinary.ID) + ' ' + MessageBinary.FileName, True);
            Index := List.Find(MessageBinary.ID);
            if (Index > - 1)
            AND (Index < List.Count)
            then begin
                List.Delete(Index);
            end;
            {$IFNDEF NEXTGEN}
            FreeAndNil(MessageBinary);
            {$ENDIF}
        end;
    end;
end;

function TTCPClientThread.CheckToReceive: Boolean;
var
    ServerResponse: String;
    Size: Int64;
    MessageBinaryID: Integer;
    MessageBinaryIndex: Integer;
    MessageBinary: TMessageBinary;
    Command: TStringList;
    EstimatedTime: TDateTime;
    ElapsedTime: TDateTime;
    MessageTransferRequest: TMessageTransferRequest;
    ChatText: String;
    Handled: Boolean;
    MessageCommand: TMessageCommand;
    i: Integer;
    //DataSizeToSend: Integer;
    BinaryPacketID: Integer;
    BinaryPacketIndex: Integer;
    TransferRate: Cardinal;
    MessageChat: TMessageChat;
    DoReceiveBinary: Boolean;
    TempReceiveStream: TMemoryStream;
begin
    Result := False;

    if NOT IdTCPClient.Connected then begin
        Exit;
    end;
    MessageBinary := nil;
    Handled := False;
    Command := TStringList.Create;
    try
        Command.Delimiter := ' ';
        try
            ServerResponse := IdTCPClient.Socket.ReadLn('', 1, - 1, ByteEncoding);
            if ServerResponse <> '' then begin
                Result := True;
                ServerResponse := Trim(ServerResponse);

                //* Send the command to the component
                if Assigned(FOnCommandReceive) then begin
                    FOnCommandReceive(Parent, IdTCPClient, ServerResponse, ByteEncoding, Handled);
                end;

                if NOT Handled then begin

                    //* Keepalive
                    if SameText(ServerResponse, 'PING') then begin
                        IdTCPClient.Socket.WriteLn('PONG', ByteEncoding);
                        Exit;
                    end;

                    Command.DelimitedText := ServerResponse;

                    //* Very bad, should not happen
                    if Command.Count <= 1 then begin
                        IdTCPClient.Disconnect;
                        Exit;
                    end;

                    //* Decide reply and action
                    //* Chat arrived
                    if SameText(Command[1], '111') then begin
                        ChatText := Copy(ServerResponse, Pos('111', ServerResponse) + 4, Length(ServerResponse));
                        //Log(TimeToStr(Now) + ' ' + ChatText, True);
                        MessageChat := TMessageChat.Create;
                        MessageChat.Text := ChatText;
                        IncomingMessageQueue.Add(MessageChat);
                        Exit;
                    end;

                    //* Very bad, should not happen
                    //if Command.Count <= 2 then begin
                    //    IdTCPClient.Disconnect;
                    //    Exit;
                    //end;

                    //* Transfer request arrived
                    if SameText(Command[1], '800') then begin
                        MessageTransferRequest := TMessageTransferRequest.Create;
                        MessageTransferRequest.DataType := StrToIntDef(Command[2], 0);
                        MessageTransferRequest.ID := StrToIntDef(Command[0], 0);
                        MessageTransferRequest.TotalSize := StrToIntDef(Command[3], 0);
                        MessageTransferRequest.StartPosition := StrToIntDef(Command[4], 0);
                        MessageTransferRequest.FileName := Copy(ServerResponse, Pos('"', ServerResponse) + 1, PosEx('"', ServerResponse, Pos('"', ServerResponse) + 1) - Pos('"', ServerResponse) - 1);
                        MessageTransferRequest.Arguments := Copy(ServerResponse, PosEx('"', ServerResponse, Pos('"', ServerResponse) + 1) + 2, Length(ServerResponse));
                        IncomingMessageQueue.Add(MessageTransferRequest);
                        Exit;
                    end;

                    //* Transfer is in progress
                    if SameText(Command[1], '802') then begin
                        MessageBinaryID := StrToIntDef(Command[0], - 1);
                        Size := StrToInt64(Trim(IdTCPClient.Socket.ReadLn(ByteEncoding)));
                        DownloadList.Lock;
                        try
                            MessageBinaryIndex := DownloadList.Find(MessageBinaryID);
                            //* Asynchronous and/or already started synchronous
                            if MessageBinaryIndex > - 1 then begin
                                MessageBinary := DownloadList.List[MessageBinaryIndex];
                                DoReceiveBinary := True;
                            //* New synchronous without agreement or client list
                            end else begin
                                if Command.Count > 2 then begin
                                    MessageBinary := TMessageBinaryMemory.Create;
                                    MessageBinary.ID := MessageBinaryID;
                                    MessageBinary.DataType := StrToIntDef(Command[2], - 1);
                                    MessageBinary.TotalSize := StrToIntDef(Command[3], 0);
                                    //MessageBinary.StartPosition := StrToIntDef(Command[4], 0);
                                    MessageBinary.FileName := Copy(ServerResponse, Pos('"', ServerResponse) + 1, PosEx('"', ServerResponse, Pos('"', ServerResponse) + 1) - Pos('"', ServerResponse) - 1);
                                    MessageBinary.Arguments := Copy(ServerResponse, PosEx('"', ServerResponse, Pos('"', ServerResponse) + 1) + 2, Length(ServerResponse));
                                    MessageBinaryIndex := DownloadList.Add(MessageBinary);
                                    DoReceiveBinary := True;
                                end else begin
                                    DoReceiveBinary := False;
                                end;
                            end;
                            if DoReceiveBinary then begin
                                if MessageBinary.DownloadStartTime = 0 then begin
                                    MessageBinary.DownloadStartTime := Now;
                                    MessageBinary.DownloadProgressMax := MessageBinary.TotalSize - MessageBinary.StartPosition;
                                    MessageBinary.DownloadLastReport := Now;
                                    ReportProgress(0, MessageBinaryID, 0, 0, 0);
                                    Log('Download started: ' + IntToStr(MessageBinaryID) + ' ' + MessageBinary.FileName, True);
                                end;
                                try
                                    TempReceiveStream := TMemoryStream.Create;
                                    try
                                        IdTCPClient.Socket.ReadStream(TempReceiveStream, Size);
                                        MessageBinary.Data.CopyFrom(TempReceiveStream, 0);
                                    finally
                                        FreeAndNil(TempReceiveStream);
                                    end;
                                except
                                    On E: Exception do begin
                                        Log('Exception while receiving binary stream: ' + E.Message, True);
                                    end;
                                end;
                                with MessageBinary do begin
                                    LastTouchTime := Now;
                                    //* Report every 250ms
                                    if DateTimeDiffmSec(DownloadLastReport, Now) > 250 then begin
                                        DownloadProgress := Round((MessageBinary.Data.Size / DownloadProgressMax) * 100);
                                        EstimatedTime := ((Now - DownloadStartTime) * DownloadProgressMax) / MessageBinary.Data.Size;
                                        ElapsedTime := Now - DownloadStartTime;
                                        if ElapsedTime <> 0 then begin
                                            TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition) / SecondSpan(ElapsedTime, 0));
                                        end else begin
                                            TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
                                        end;
                                        ReportProgress(DownloadProgress, MessageBinaryID, ElapsedTime, EstimatedTime, TransferRate);
                                        DownloadLastReport := Now;
                                    end;
                                end;
                                if MessageBinary.Data.Size = MessageBinary.TotalSize then begin
                                    DownloadList.Delete(MessageBinaryIndex);
                                    MessageBinary.DownloadProgress := 100;
                                    ElapsedTime := Now - MessageBinary.DownloadStartTime;
                                    if ElapsedTime <> 0 then begin
                                        TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition) / SecondSpan(ElapsedTime, 0));
                                    end else begin
                                        TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
                                    end;
                                    MessageBinary.Data.Seek(0, soBeginning);
                                    ReportProgress(100, MessageBinaryID, ElapsedTime, 0, TransferRate);
                                    Log('Download completed: ' + IntToStr(MessageBinary.ID) + ' ' + MessageBinary.FileName, True);
                                    //if MessageBinary.DataType = TCP_SOCKET_DATA_TYPE_CLIENT_LIST then begin
                                    // ProcessIncomingClientList(MessageBinary);
                                    //    FreeAndNil(MessageBinary);
                                    //end else begin
                                        IncomingMessageQueue.Add(MessageBinary);
                                    //end;
                                end;
                            end else begin
                                IdTCPClient.IOHandler.Discard(Size);
                            end;
                        finally
                            DownloadList.UnLock;
                        end;
                        Exit;
                    end;
                //end;

                    //* Server sent a command
                    if SameText(Command[1], '700')
                    then begin
                        MessageCommand := TMessageCommand.Create;
                        MessageCommand.ID := StrToIntDef(Command[0], 0);
                        MessageCommand.Command := Command[2];
                        MessageCommand.Arguments := Copy(ServerResponse, Pos(Command[2], ServerResponse) + Length(Command[2]) + 1, Length(ServerResponse));
                        MessageCommand.Client := nil;
                        IncomingMessageQueue.Add(MessageCommand);
                        Exit;
                    end;

                    //* Data transfer agreed
                    if SameText(Command[1], '200')
                    AND SameText(Command[2], '801')
                    then begin
                        BinaryPacketID := StrToIntDef(Command[0], 0);
                        if NOT DataTransferAgreed(BinaryPacketID, SendList) then begin
                            DataTransferAgreed(BinaryPacketID, SendListSynchronous);
                        end;
                        Exit;
                    end;

                    //* Data transfer cancel
                    if SameText(Command[1], '809') then begin
                        BinaryPacketID := StrToIntDef(Command[0], 0);
                        if NOT DataTransferCancel(BinaryPacketID, SendList) then begin
                            if NOT DataTransferCancel(BinaryPacketID, SendListSynchronous) then begin
                                DownloadList.Lock;
                                try
                                    BinaryPacketIndex := DownloadList.Find(BinaryPacketID);
                                    if BinaryPacketIndex > - 1 then begin
                                        MessageBinary := TMessageBinary(DownloadList.List[BinaryPacketIndex]);
                                        DownloadList.Delete(BinaryPacketIndex);
                                        IncomingMessageQueue.Add(MessageBinary);
                                        Log('Binary transfer canceled: ' + IntToStr(BinaryPacketID), True);
                                    end;
                                finally
                                    DownloadList.UnLock;
                                end;
                            end;
                        end;
                        Exit;
                    end;
                end;
            end;

            //* Is there binary to send? (Asynchronous)
            SendList.Lock;
            try
                i :=  - 1;
                while i <= SendList.Count - 2 do begin
                    Inc(i);
                    if TObject(SendList.List[i]) is TMessageBinary then begin
                        DoBinaryTransfer(IdTCPClient, TMessageBinary(SendList.List[i]), SendList);
                    end;
                end;
            finally
                SendList.UnLock;
            end;

            //* Is there binary to send? (Synchronous)
            SendListSynchronous.Lock;
            try
                if SendListSynchronous.Count > 0 then begin
                    if TObject(SendListSynchronous.List[0]) is TMessageBinary then begin
                        DoBinaryTransfer(IdTCPClient, TMessageBinary(SendListSynchronous.List[0]), SendListSynchronous);
                    end;
                end;
            finally
                SendListSynchronous.UnLock;
            end;

        except
            On E: exception do begin
                Log('Exception: ' + E.Message, True);
            end;
        end;
    finally
        FreeAndNil(Command);
    end;
end;

procedure TTCPClientThread.IdTCPClientConnected(Sender: TObject);
var
    ServerResponse: String;
    SendPassWord: String;
    MessageConnected: TMessageConnected;
    MessageLogIn: TMessageLogIn;
begin
    IdTCPClient.Socket.SendBufferSize := FParameters.DataTransferPacketSize;
    ServerResponse := IdTCPClient.Socket.ReadLn(ByteEncoding);
    Log('Connected to server: ' + Trim(ServerResponse), True);
    ServerResponse := IdTCPClient.Socket.ReadLn(ByteEncoding);
    if Trim(ServerResponse) = '511 Login name?' then begin
        IdTCPClient.Socket.WriteLn(LoginName, ByteEncoding);
        ServerResponse := IdTCPClient.Socket.ReadLn(ByteEncoding);
        case FLoginHashingMode of
            hmOpenSSL: begin
                SendPassWord := HashSHA512String(Trim(ServerResponse) + PassWord);
            end;
            hmMD5: begin
                SendPassWord := HashMD5String(Trim(ServerResponse) + PassWord);
            end;
        end;
        IdTCPClient.Socket.WriteLn(SendPassWord, ByteEncoding);
        ServerResponse := IdTCPClient.Socket.ReadLn(ByteEncoding);
        if Trim(ServerResponse) = '200 OK' then begin
            Log('Login accepted. Succesfully connected.', True);
            MessageConnected := TMessageConnected.Create;
            MessageConnected.Connected := True;
            IncomingMessageQueue.Add(MessageConnected);
            ServerResponse := IdTCPClient.Socket.ReadLn(ByteEncoding);
            if Trim(ServerResponse) = '100 User name?' then begin
                IdTCPClient.Socket.WriteLn(UserName, ByteEncoding);
                ServerResponse := IdTCPClient.Socket.ReadLn(ByteEncoding);
                if SameText(Trim(ServerResponse), '200 OK') then begin
                    Log('Succesfully set user name: ' + UserName, True);
                end
            end;
            if Assigned(FOnConnected) then begin
                FOnConnected(Parent, IdTCPClient, ByteEncoding);
            end;
        end else begin
            ErrorIncorrectLogIn := True;
            MessageLogIn := TMessageLogIn.Create;
            MessageLogIn.Successfull := False;
            IncomingMessageQueue.Add(MessageLogIn);
            Log('Login rejected.', True);
            IdTCPClient.Disconnect;
        end;
    end else begin
        IdTCPClient.Disconnect;
    end;
end;

procedure TTCPClientThread.IdTCPClientDisconnected(Sender: TObject);
var
    MessageConnected: TMessageConnected;
begin
    if Assigned(FOnDisconnected) then begin
        FOnDisconnected(Parent, IdTCPClient);
    end;
    MessageConnected := TMessageConnected.Create;
    MessageConnected.Connected := False;
    IncomingMessageQueue.Add(MessageConnected);
    Log('Disconnected from server.', True);
end;

procedure TTCPClientThread.LockTransferLists;
begin
    DownloadList.Lock;
    SendList.Lock;
    SendListSynchronous.Lock;
end;

procedure TTCPClientThread.Log(const Text: String; Debug: Boolean);
var
    MessageLog: TMessageLog;
begin
    if NOT LoggingOn then begin
        Exit;
    end;
    MessageLog := TMessageLog.Create;
    MessageLog.Text := Text;
    MessageLog.Debug := Debug;
    IncomingMessageQueue.Add(MessageLog);
end;

procedure TTCPClientThread.ReportProgress(Progress: Integer; TransferID: Integer; ElapsedTime, EstimatedTime: TDateTime; TransferRate: Cardinal);
var
    MessageProgress: TMessageProgress;
begin
    MessageProgress := TMessageProgress.Create;
    MessageProgress.ID := TransferID;
    MessageProgress.Progress := Progress;
    MessageProgress.ElapsedTime := ElapsedTime;
    MessageProgress.EstimatedTime := EstimatedTime;
    if EstimatedTime = 0 then begin
        MessageProgress.RemainingTime := 0;
    end else begin
        MessageProgress.RemainingTime := EstimatedTime - ElapsedTime;
    end;
    MessageProgress.TransferRate := TransferRate;
    IncomingMessageQueue.Add(MessageProgress);
end;

procedure TTCPClientThread.SendChat(Chat: TMessageChat);
begin
    try
        if Chat.Text <> '' then begin
            IdTCPClient.Socket.WriteLn('0 111 ' + Chat.Text, ByteEncoding);
        end;
    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(Chat);
        {$ENDIF}
    end;
end;

procedure TTCPClientThread.SendCommand(MessageCommand: TMessageCommand);
begin
    try
        if MessageCommand.Command <> '' then begin
            if MessageCommand.ID = 0 then begin
                MessageCommand.ID := GenerateUniqueCommandID;
            end;
            IdTCPClient.Socket.WriteLn(IntToStr(MessageCommand.ID) + ' 700 ' + MessageCommand.Command + ' ' + MessageCommand.Arguments, ByteEncoding); //send string to server
            CommandList.Lock;
            try
                CommandList.Add(MessageCommand);
            finally
                CommandList.UnLock;
            end;
        end;
    finally
        //*
    end;
end;

procedure TTCPClientThread.SendBinary(MessageBinary: TMessageBinary);
begin
    try
        if MessageBinary.Data.Size > 0 then begin
            if MessageBinary.Arguments = '' then begin
                MessageBinary.Arguments := '0';
            end;
            if MessageBinary.ID = 0 then begin
                LockTransferLists;
                try
                    MessageBinary.ID := GenerateUniqueTransferID;
                finally
                    UnLockTransferLists;
                end;
            end;
            case MessageBinary.TransferType of
                ttAsynchronous: begin
                    //* Send request
                    if MessageBinary.RequireAgreement = raTrue then begin
                        IdTCPClient.Socket.WriteLn(IntToStr(MessageBinary.ID) + ' 800 ' + IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size) + ' "' + MessageBinary.FileName + '"' + ' ' + MessageBinary.Arguments, ByteEncoding);
                        MessageBinary.RequestSent := True;
                    end;
                    //* Start transfer instantly
                    if MessageBinary.RequireAgreement = raFalse then begin
                        MessageBinary.RequestSent := True;
                        MessageBinary.DoTransfer := True;
                    end;
                    SendList.Lock;
                    try
                        SendList.Add(MessageBinary);
                    finally
                        SendList.UnLock;
                    end;
                end;
                ttSynchronous: begin
                    SendListSynchronous.Lock;
                    try
                        //* Send request
                        if MessageBinary.RequireAgreement = raTrue then begin
                            IdTCPClient.Socket.WriteLn(IntToStr(MessageBinary.ID) + ' 800 ' + IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size) + ' "' + MessageBinary.FileName + '"' + ' ' + MessageBinary.Arguments, ByteEncoding);
                            MessageBinary.RequestSent := True;
                        end;
                        //* Start transfer instantly
                        if MessageBinary.RequireAgreement = raFalse then begin
                            MessageBinary.RequestSent := True;
                            MessageBinary.DoTransfer := True;
                        end;
                        SendListSynchronous.Add(MessageBinary);
                    finally
                        SendListSynchronous.UnLock;
                    end;
                end;
            end;
        end;
    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(MessageBinary);
        {$ENDIF}
    end;
end;

procedure TTCPClientThread.ProcessTransferRequest(MessageTransferRequest: TMessageTransferRequest);
var
    MessageBinary: TMessageBinary;
begin
    try
        if MessageTransferRequest.Accept then begin
            if MessageTransferRequest.SaveToFile then begin
                MessageBinary := TMessageBinaryFile.Create(MessageTransferRequest.SaveToFileName, MessageTransferRequest.Resume);
            end else begin
                MessageBinary := TMessageBinaryMemory.Create;
            end;
            MessageBinary.DataType := MessageTransferRequest.DataType;
            MessageBinary.ID := MessageTransferRequest.ID;
            MessageBinary.Arguments := MessageTransferRequest.Arguments;
            MessageBinary.TotalSize := MessageTransferRequest.TotalSize;
            MessageBinary.StartPosition := MessageTransferRequest.StartPosition;
            MessageBinary.FileName := MessageTransferRequest.FileName;
            MessageBinary.User := MessageTransferRequest.User;
            DownloadList.Lock;
            try
                DownloadList.Add(MessageBinary);
            finally
                DownloadList.UnLock;
            end;
            IdTCPClient.Socket.WriteLn(IntToStr(MessageBinary.ID) + ' 200 801', ByteEncoding);
        end;
    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(MessageTransferRequest);
        {$ENDIF}
    end;
end;

procedure TTCPClientThread.ProcessTransferCancel(MessageTransferCancel: TMessageTransferCancel);
var
    Index: Integer;
begin
    try
        //* Send a cancel request/notification
        IdTCPClient.Socket.WriteLn(IntToStr(MessageTransferCancel.ID) + ' 809', ByteEncoding);
        //* Remove download

        DownloadList.Lock;
        try
            Index := DownloadList.Find(MessageTransferCancel.ID);
            if Index > - 1 then begin
                {$IFNDEF NEXTGEN}
                TMessageBinary(DownloadList.List[Index]).Free;
                {$ENDIF}
                DownloadList.Delete(Index);
            end;
        finally
            DownloadList.UnLock;
        end;

        SendList.Lock;
        try
            Index := SendList.Find(MessageTransferCancel.ID);
            if Index > - 1 then begin
                {$IFNDEF NEXTGEN}
                TMessageBinary(SendList.List[Index]).Free;
                {$ENDIF}
                SendList.Delete(Index);
            end;
        finally
            SendList.UnLock;
        end;

        SendListSynchronous.Lock;
        try
            Index := SendListSynchronous.Find(MessageTransferCancel.ID);
            if Index > - 1 then begin
                {$IFNDEF NEXTGEN}
                TMessageBinary(SendListSynchronous.List[Index]).Free;
                {$ENDIF}
                SendListSynchronous.Delete(Index);
            end;
        finally
            SendListSynchronous.UnLock;
        end;

    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(MessageTransferCancel);
        {$ENDIF}
    end;
end;

procedure TTCPClientThread.UnLockTransferLists;
begin
    DownloadList.UnLock;
    SendList.UnLock;
    SendListSynchronous.UnLock;
end;

end.
