//********************************************************************************************************************************
//*                                                                                                                              *
//*     TCP Server-Client Library 1.0.40.80  3delite 2013-2025                                                                  *
//*     See TCP Server-Client Library ReadMe.txt for details.                                                                    *
//*                                                                                                                              *
//* Licenses available for usage of this component:                                                                              *
//* Freeware License: 25                                                                                                        *
//* Shareware License: 50                                                                                                       *
//* Commercial License: 250                                                                                                     *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/register.html#tcpscl                                      *
//*                                                                                                                              *
//* Home page:                                                                                                                   *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/TCPServerClientLibrary.html                               *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     https://www.3delite.hu/                                                                                                  *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit TCPLibraryClasses;

interface

{$I TCPSCLCompilerDefines.inc}

Uses
    SysUtils,
    SyncObjs,
    Types,
    {$IFDEF TCPSCL_GENERICS}
    Generics.Collections,
    {$ENDIF}
    Classes,
    IdGlobal,
    IdContext,
    IdTCPClient,
    UsersList,
    TCPSCLClientList;

Const
    //* Binary transfer data types
    TCP_SOCKET_DATA_TYPE_CLIENT_LIST    = 1000;
    //* Examples
    TCP_SOCKET_DATA_TYPE_PICTURE        = 1;
    TCP_SOCKET_DATA_TYPE_FILE           = 2;

Const
    MAX_PROTOCOL_LEVEL              = 1;
    DATA_TRANSFER_PACKET_SIZE       = 1024 * 32;        //* 32 KBytes
    CLIENT_TIMEOUT_DELAY            = 1000 * 60 * 5;    //* 5 minutes
    CLIENT_PING_INTERVAL            = 1000 * 60 * 1;    //* 1 minutes
    TRANSFER_DATA_PURGE_INTERVAL    = 1000 * 60 * 1;    //* 1 minutes
    TRANSFER_DATA_INVALID_INTERVAL  = 1000 * 60 * 5;    //* 5 minutes

type
    TTCPServerParams = class (TPersistent)
    private
        FTransferDataInvalidInterval: Integer;
        FClientTimeoutDelay: Integer;
        FDataTransferPacketSize: Integer;
        FClientPingInterval: Integer;
        FTransferDataPurgeInterval: Integer;
    public
        Constructor Create;
    published
        property TransferDataInvalidInterval: Integer read FTransferDataInvalidInterval write FTransferDataInvalidInterval default TRANSFER_DATA_INVALID_INTERVAL;
        property ClientTimeoutDelay: Integer read FClientTimeoutDelay write FClientTimeoutDelay default CLIENT_TIMEOUT_DELAY;
        property DataTransferPacketSize: Integer read FDataTransferPacketSize write FDataTransferPacketSize default DATA_TRANSFER_PACKET_SIZE;
        property ClientPingInterval: Integer read FClientPingInterval write FClientPingInterval default CLIENT_PING_INTERVAL;
        property TransferDataPurgeInterval: Integer read FTransferDataPurgeInterval write FTransferDataPurgeInterval default TRANSFER_DATA_PURGE_INTERVAL;
    end;

type
    TTCPClientParams = class (TPersistent)
    private
        FTransferDataInvalidInterval: Integer;
        FDataTransferPacketSize: Integer;
        FTransferDataPurgeInterval: Integer;
    public
        Constructor Create;
    published
        property TransferDataInvalidInterval: Integer read FTransferDataInvalidInterval write FTransferDataInvalidInterval default TRANSFER_DATA_INVALID_INTERVAL;
        property DataTransferPacketSize: Integer read FDataTransferPacketSize write FDataTransferPacketSize default DATA_TRANSFER_PACKET_SIZE;
        property TransferDataPurgeInterval: Integer read FTransferDataPurgeInterval write FTransferDataPurgeInterval default TRANSFER_DATA_PURGE_INTERVAL;
    end;

type
    TTransferType = (ttAsynchronous, ttSynchronous);
    TRequireAgreement = (raTrue, raFalse);
    TTCPSCLHashingMode = (hmOpenSSL, hmMD5);

type
    TTCPServerMessage = class {abstract}
        {$IFDEF NEXTGEN}[Weak] {$ENDIF}Context: TIdContext;
    end;

    TMessageConnected = class(TTCPServerMessage)
        Connected: Boolean;
    end;

    TMessageTerminated = class(TTCPServerMessage)
        Terminated: Boolean;
    end;

    TMessageLogIn = class(TTCPServerMessage)
        Successfull: Boolean;
    end;

    TMessageLog = class(TTCPServerMessage)
        Text: String;
        Debug: Boolean;
    end;

    TTCPServerMessageID = class {abstract} (TTCPServerMessage)
        ID: Integer;
    end;

    TMessageProgress = class(TTCPServerMessageID)
        Progress: Integer;
        ElapsedTime: TDateTime;
        EstimatedTime: TDateTime;
        RemainingTime: TDateTime;
        TransferRate: Cardinal;
    end;

    TMessageTransferRequest = class(TTCPServerMessageID)
        DataType: Integer;
        Arguments: String;
        FileName: String;
        TotalSize: Integer;
        Resume: Boolean;
        StartPosition: Int64;
        Accept: Boolean;
        SaveToFile: Boolean;
        SaveToFileName: String;
        User: TObject;
    end;

    TMessageTransferCancel = class(TTCPServerMessageID)
    end;

    TMessageBinary = class {abstract} (TTCPServerMessageID)
        FileName: String;
        DataType: Integer;
        Arguments: String;
        Data: TStream;
        TotalSize: Int64;
        Resume: Boolean;
        StartPosition: Int64;
        DownloadStartTime: TDateTime;
        DownloadProgress: Int64;
        DownloadProgressMax: Int64;
        DownloadLastReport: TDateTime;
        LastTouchTime: TDateTime;
        RequestSent: Boolean;
        DoTransfer: Boolean;
        TransferType: TTransferType;
        RequireAgreement: TRequireAgreement;
        User: TObject;
        Constructor Create;
        Destructor Destroy; override;
    end;

    TMessageBinaryMemory = class(TMessageBinary)
        Constructor Create; overload;
        Constructor Create(Text: String); overload;
        Constructor Create(Strings: TStrings); overload;
        Constructor Create(Stream: TStream); overload;
    end;

    TMessageBinaryFile = class(TMessageBinary)
        LocalFileName: String;
        Constructor Create(FileName: String; _Resume: Boolean = False);
    end;

    TMessageChat = class(TTCPServerMessageID)
        UserName: String;
        Text: String;
    end;

    TMessageClientKick = class
        {$IFDEF NEXTGEN}[Weak] {$ENDIF}Context: TIdContext;
    end;

    TDownloadList = class
    private
        FCriticalSection: TCriticalSection;
    public
        {$IFDEF TCPSCL_GENERICS}
        List: TList<TMessageBinary>;
        {$ELSE}
        List: TList;
        {$ENDIF}
        Constructor Create;
        Destructor Destroy; override;
        procedure Clear;
        function Count: Integer;
        function Add(Item: TMessageBinary): Integer;
        function Find(ID: Integer): Integer;
        procedure Delete(Index: Integer);
        procedure Lock;
        procedure UnLock;
    end;

    TSendList = class
    private
        FCriticalSection: TCriticalSection;
    public
        {$IFDEF TCPSCL_GENERICS}
        List: TList<TTCPServerMessageID>;
        {$ELSE}
        List: TList;
        {$ENDIF}
        Constructor Create;
        Destructor Destroy; override;
        procedure Clear;
        function Count: Integer;
        procedure Add(Item: TTCPServerMessageID);
        function Find(ID: Integer): Integer;
        procedure Delete(Index: Integer);
        procedure Lock;
        procedure UnLock;
    end;

    TMessageCommand = class;

    TCommandList = class
    private
        FCriticalSection: TCriticalSection;
    public
        {$IFDEF TCPSCL_GENERICS}
        List: TList<TMessageCommand>;
        {$ELSE}
        List: TList;
        {$ENDIF}
        Constructor Create;
        Destructor Destroy; override;
        procedure Clear;
        function Count: Integer;
        procedure Add(Item: TMessageCommand){: TMessageCommand};
        function Find(ID: Integer): Integer;
        procedure Delete(Index: Integer);
        procedure Lock;
        procedure UnLock;
    end;

    TConnectionInfo = class
        LastHeartBeatDateTime: TDateTime;
    end;

    TConnectedClient = class
        {$IFDEF NEXTGEN}[Weak] {$ENDIF}Context: TIdContext;
        LoginName: String;
        UserName: String;
        ProtocolLevel: Cardinal;
        SendList: TSendList;
        SendListSynchronous: TSendList;
        UserRights: TUserRights;
        IPAddress: String;
        Data: TObject;
        DownloadList: TDownloadList;
        CommandList: TCommandList;
        ID: TGUID;
        ExpectsClientList: Boolean;
        function GenerateUniqueTransferID: Integer;
        function GenerateUniqueCommandID: String;
        procedure LockTransferLists;
        procedure UnLockTransferLists;
        Constructor Create;
        Destructor Destroy; override;
    end;

    TConnectedClients = class
    private
        FCriticalSection: TCriticalSection;
    public
        {$IFDEF TCPSCL_GENERICS}
        List: TList<TConnectedClient>;
        {$ELSE}
        List: TList;
        {$ENDIF}
        Constructor Create;
        Destructor Destroy; override;
        function Add(Context: TIdContext; LoginName, UserName: String; ProtocolLevel: Cardinal; UserRights: TUserRights): Boolean;
        function Remove(Context: TIdContext): Boolean; overload;
        function Remove(ID: TGUID): Boolean; overload;
        function Find(Context: TIdContext): TConnectedClient; overload;
        function Find(ID: TGUID): TConnectedClient; overload;
        procedure Lock;
        procedure UnLock;
        function GenerateClientList(ClientList: TStringList): Boolean;
    end;

    TMessageCommand = class(TTCPServerMessageID)
        Command: String;
        Arguments: String;
        Client: TConnectedClient;
        Data: TObject;
        Destructor Destroy; override;
    end;

type
    TServerCommandReceiveEvent = procedure(Sender: TObject; AContext: TIdContext; Client: TConnectedClient; Command: String; {$IFDEF INDY_ENCODING} Encoding: IIdTextEncoding; {$ELSE} Encoding: TEncoding; {$ENDIF} var Handled: Boolean) of object;
    TClientCommandReceiveEvent = procedure(Sender: TObject; IdTCPClient: TIdTCPClient; Command: String; {$IFDEF INDY_ENCODING} Encoding: IIdTextEncoding; {$ELSE} Encoding: TEncoding; {$ENDIF} var Handled: Boolean) of object;
    TServerClientConnectedEvent = procedure(Sender: TObject; AContext: TIdContext; Client: TConnectedClient; {$IFDEF INDY_ENCODING} Encoding: IIdTextEncoding {$ELSE} Encoding: TEncoding {$ENDIF}) of object;
    TClientConnectedEvent = procedure(Sender: TObject; IdTCPClient: TIdTCPClient; {$IFDEF INDY_ENCODING} Encoding: IIdTextEncoding {$ELSE} Encoding: TEncoding {$ENDIF}) of object;
    TServerClientDisconnectedEvent = procedure(Sender: TObject; AContext: TIdContext; Client: TConnectedClient) of object;
    TClientDisconnectedEvent = procedure(Sender: TObject; IdTCPClient: TIdTCPClient) of object;
    TServerLoginQueryEvent = procedure(Sender: TObject; const LoginName: String; var PassWord: String; var Deny: Boolean) of object;
    TServerBeforeActivateEvent = procedure(Sender: TObject) of object;
    TServerAfterActivateEvent = procedure(Sender: TObject) of object;
    TServerBeforeDeActivateEvent = procedure(Sender: TObject) of object;
    TServerAfterDeActivateEvent = procedure(Sender: TObject) of object;
    TServerThreadStartEvent = procedure(Sender: TObject) of object;
    TServerThreadEndEvent = procedure(Sender: TObject) of object;
    TServerCustomSendEvent = procedure(Sender: TObject; AContext: TIdContext; Client: TConnectedClient; {$IFDEF INDY_ENCODING} Encoding: IIdTextEncoding {$ELSE} Encoding: TEncoding {$ENDIF}) of object;
    TClientCustomSendEvent = procedure(Sender: TObject; IdTCPClient: TIdTCPClient; {$IFDEF INDY_ENCODING} Encoding: IIdTextEncoding; {$ELSE} Encoding: TEncoding; {$ENDIF} MessageObject: TObject) of object;
    TServerResultMessageEvent = procedure(Sender: TObject; TCPServerMessage: TTCPServerMessage) of object;
    TClientResultMessageEvent = procedure(Sender: TObject; TCPClientMessage: TTCPServerMessage) of object;
    TClientConnectionChangeEvent = procedure(Sender: TObject; Client: TClient) of object;

    function DateTimeDiff(Start, Stop: TDateTime): Int64;
    function DateTimeDiffmSec(Start, Stop: TDateTime): Int64;
    function Sec2Time(Sec: Int64): String;

implementation

function DateTimeDiff(Start, Stop: TDateTime): Int64;
var
    TimeStamp : TTimeStamp;
begin
    TimeStamp := DateTimeToTimeStamp(Stop - Start);
    Dec(TimeStamp.Date, TTimeStamp(DateTimeToTimeStamp(0)).Date);
    Result := (TimeStamp.Date * 24 * 60 * 60) + (TimeStamp.Time div 1000);
end;

function DateTimeDiffmSec(Start, Stop: TDateTime): Int64;
var
    TimeStamp : TTimeStamp;
begin
    TimeStamp := DateTimeToTimeStamp(Stop - Start);
    Dec(TimeStamp.Date, TTimeStamp(DateTimeToTimeStamp(0)).Date);
    Result := (TimeStamp.Date * 24 * 60 * 60 * 1000) + (TimeStamp.Time);
end;

function Sec2Time(Sec: Int64): String;
var
    tDays: String;
    tHours: String;
    tMinutes: String;
    tSecs: String;
    //tmSecs: String;
    Seconds: Int64;
    mSec: Int64;
begin
    mSec := Sec * 1000;
    if mSec = 4294967295
        then mSec := 0;
    Seconds := mSec div 1000;
    tDays := IntToStr(Seconds div 86400);
    tHours := IntToStr((Seconds - (StrToInt(tDays) * 86400)) div 3600);
    if Length(tHours) = 1
        then tHours := '0' + tHours;
    tMinutes := IntToStr(((Seconds - (StrToInt(tDays) * 86400) - (StrToInt(tHours) * 3600)) div 60));
    if Length(tMinutes) = 1
        then tMinutes := '0' + tMinutes;
    tSecs := IntToStr(Seconds - (StrToInt(tDays) * 86400) - (StrToInt(tHours) * 3600 + StrToInt(tMinutes) * 60));
    if Length(tSecs) = 1
        then tSecs := '0' + tSecs;
    {
    tmSecs := IntToStr(mSec - (((StrToInt(tDays) * 86400) + StrToInt(tHours) * 3600 + StrToInt(tMinutes) * 60) + StrToInt(tSecs)) * 1000);
    case Length(tmSecs) of
        1: tmSecs := '00' + tmSecs;
        2: tmSecs := '0' + tmSecs;
    end;
    tmSecs := Copy(tmSecs, 1, 2);
    }
    if tDays <> '0' then begin
        if tDays = '1' then begin
            Result := tDays + ' day ' + tHours + ':' + tMinutes + ':' + tSecs{ + '.' + tmSecs};
        end else begin
            Result := tDays + ' days ' + tHours + ':' + tMinutes + ':' + tSecs{ + '.' + tmSecs};
        end;
    end else begin
        Result := tHours + ':' + tMinutes + ':' + tSecs{ + '.' + tmSecs};
    end;
end;

{ TConnectedClients }

function TConnectedClients.Add(Context: TIdContext; LoginName, UserName: String; ProtocolLevel: Cardinal; UserRights: TUserRights): Boolean;
var
    ConnectedClient: TConnectedClient;
    i: Integer;
    GUIDValid: Boolean;
begin
    ConnectedClient := TConnectedClient.Create;
    ConnectedClient.Context := Context;
    ConnectedClient.LoginName := LoginName;
    ConnectedClient.UserName := UserName;
    ConnectedClient.ProtocolLevel := ProtocolLevel;
    ConnectedClient.UserRights := UserRights;
    ConnectedClient.IPAddress := Context.Binding.PeerIP;
    //* Check if same GUID exists, create new GUID if needed and add client to the list
    Self.Lock;
    try
        repeat
            GUIDValid := True;
            for i := 0 to List.Count - 1 do begin
                if List[i].ID = ConnectedClient.ID then begin
                    GUIDValid := False;
                    Break;
                end;
            end;
            if NOT GUIDValid then begin
                CreateGUID(ConnectedClient.ID);
            end;
        until GUIDValid;
        List.Add(ConnectedClient);
    finally
        Self.UnLock;
    end;
    Result := True;
end;

constructor TConnectedClients.Create;
begin
    inherited;
    {$IFDEF TCPSCL_GENERICS}
    List := TList<TConnectedClient>.Create;
    {$ELSE}
    List := TList.Create;
    {$ENDIF}
    FCriticalSection := TCriticalSection.Create;
end;

destructor TConnectedClients.Destroy;
begin
    FreeAndNil(List);
    FreeAndNil(FCriticalSection);
    inherited;
end;

function TConnectedClients.Find(Context: TIdContext): TConnectedClient;
var
    i: Integer;
begin
    Result := nil;
    for i := 0 to List.Count - 1 do begin
        if TConnectedClient(List[i]).Context = Context then begin
            Result := TConnectedClient(List[i]);
            Break;
        end;
    end;
end;

function TConnectedClients.Find(ID: TGUID): TConnectedClient;
var
    i: Integer;
begin
    Result := nil;
    for i := 0 to List.Count - 1 do begin
        if TConnectedClient(List[i]).ID = ID then begin
            Result := TConnectedClient(List[i]);
            Break;
        end;
    end;
end;

function TConnectedClients.GenerateClientList(ClientList: TStringList): Boolean;
var
    i: Integer;
begin
    Result := False;
    ClientList.Clear;
    Lock;
    try
        for i := 0 to List.Count - 1 do begin
            {$IFDEF TCPSCL_GENERICS}
            ClientList.Append(GUIDToString(List[i].ID) + ',LOGINNAME=' + List[i].LoginName + ',USERNAME=' + List[i].UserName + ',IPADDRESS=' + List[i].IPAddress);
            {$ELSE}
            ClientList.Append(GUIDToString((List[i] as TConnectedClient).ID) + ',LOGINNAME=' + (List[i] as TConnectedClient).LoginName + ',USERNAME=' + (List[i] as TConnectedClient).UserName + ',IPADDRESS=' + (List[i] as TConnectedClient).IPAddress);
            {$ENDIF}
            Result := True;
        end;
    finally
        UnLock;
    end;
end;

procedure TConnectedClients.Lock;
begin
    FCriticalSection.Enter;
end;

function TConnectedClients.Remove(Context: TIdContext): Boolean;
var
    i: Integer;
begin
    Result := False;
    for i := 0 to List.Count - 1 do begin
        if TConnectedClient(List[i]).Context = Context then begin
            {$IFNDEF NEXTGEN}
            TConnectedClient(List[i]).Free;
            {$ENDIF}
            List.Delete(i);
            Break;
        end;
    end;
end;

function TConnectedClients.Remove(ID: TGUID): Boolean;
var
    i: Integer;
begin
    Result := False;
    for i := 0 to List.Count - 1 do begin
        if TConnectedClient(List[i]).ID = ID then begin
            {$IFNDEF NEXTGEN}
            TConnectedClient(List[i]).Free;
            {$ENDIF}
            List.Delete(i);
            Break;
        end;
    end;
end;

procedure TConnectedClients.UnLock;
begin
    FCriticalSection.Leave;
end;

{ TMessageBinary }

constructor TMessageBinary.Create;
begin
    inherited;
    //Data := TMemoryStream.Create;
end;

destructor TMessageBinary.Destroy;
begin
    if Assigned(Data) then begin
        FreeAndNil(Data);
    end;
    inherited;
end;

{ TSendList }

procedure TSendList.Add(Item: TTCPServerMessageID);
var
    NewItem: TMessageBinary;
    NewItemCommand: TMessageCommand;
    ChatItem: TMessageChat;
begin
    if Item is TMessageBinaryMemory then begin
        with Item as TMessageBinaryMemory do begin
            NewItem := TMessageBinaryMemory.Create;
            NewItem.Data.CopyFrom(Data, 0);
            NewItem.Data.Seek(StartPosition, soBeginning);
            NewItem.FileName := FileName;
            NewItem.DataType := DataType;
            NewItem.ID := ID;
            NewItem.Arguments := Arguments;
            NewItem.LastTouchTime := Now;
            NewItem.TotalSize := NewItem.Data.Size;
            NewItem.RequestSent := RequestSent;
            NewItem.DoTransfer := DoTransfer;
            NewItem.TransferType := TransferType;
            NewItem.RequireAgreement := RequireAgreement;
            NewItem.User := User;
        end;
        List.Add(NewItem);
    end;
    if Item is TMessageBinaryFile then begin
        with Item as TMessageBinaryFile do begin
            NewItem := TMessageBinaryFile.Create(TMessageBinaryFile(Item).LocalFileName, TMessageBinaryFile(Item).Resume);
            NewItem.Data.Seek(StartPosition, soBeginning);
            NewItem.FileName := FileName;
            NewItem.DataType := DataType;
            NewItem.ID := ID;
            NewItem.Arguments := Arguments;
            NewItem.LastTouchTime := Now;
            NewItem.TotalSize := NewItem.Data.Size;
            NewItem.RequestSent := RequestSent;
            NewItem.DoTransfer := DoTransfer;
            NewItem.TransferType := TransferType;
            NewItem.RequireAgreement := RequireAgreement;
            NewItem.User := User;
        end;
        List.Add(NewItem);
    end;
    if Item is TMessageCommand then begin
        with Item as TMessageCommand do begin
            NewItemCommand := TMessageCommand.Create;
            NewItemCommand.ID := ID;
            NewItemCommand.Command := Command;
            NewItemCommand.Arguments := Arguments;
            NewItemCommand.Client := Client;
            NewItemCommand.Context := Context;
            NewItemCommand.Data := Data;
        end;
        List.Add(NewItemCommand);
    end;
    if Item is TMessageChat then begin
        with Item as TMessageChat do begin
            ChatItem := TMessageChat.Create;
            ChatItem.UserName := UserName;
            ChatItem.Text := (Item as TMessageChat).Text;
        end;
        List.Add(ChatItem);
    end;
end;

procedure TSendList.Clear;
var
    i: Integer;
begin
    for i := List.Count - 1 downto 0 do begin
        {$IFNDEF NEXTGEN}
        TObject(List[i]).Free;
        {$ENDIF}
        List.Delete(i);
    end;
end;

function TSendList.Count: Integer;
begin
    Result := List.Count;
end;

constructor TSendList.Create;
begin
    inherited;
    FCriticalSection := TCriticalSection.Create;
    {$IFDEF TCPSCL_GENERICS}
    List := TList<TTCPServerMessageID>.Create;
    {$ELSE}
    List := TList.Create;
    {$ENDIF}
end;

procedure TSendList.Delete(Index: Integer);
begin
    List.Delete(Index);
end;

destructor TSendList.Destroy;
begin
    Clear;
    FreeAndNil(List);
    FreeAndNil(FCriticalSection);
    inherited;
end;

function TSendList.Find(ID: Integer): Integer;
var
    i: Integer;
begin
    Result := - 1;
    for i := 0 to List.Count - 1 do begin
        if TMessageBinary(List[i]).ID = ID then begin
            Result := i;
            Break;
        end;
    end;
end;

procedure TSendList.Lock;
begin
    FCriticalSection.Enter;
end;

procedure TSendList.UnLock;
begin
    FCriticalSection.Leave;
end;

{ TConnectedClient }

constructor TConnectedClient.Create;
begin
    inherited;
    SendList := TSendList.Create;
    SendListSynchronous := TSendList.Create;
    DownloadList := TDownloadList.Create;
    CommandList := TCommandList.Create;
    CreateGUID(ID);
end;

destructor TConnectedClient.Destroy;
begin
    FreeAndNil(SendList);
    FreeAndNil(SendListSynchronous);
    FreeAndNil(DownloadList);
    FreeAndNil(CommandList);
    if Assigned(Data) then begin
        FreeAndNil(Data);
    end;
    inherited;
end;

function TConnectedClient.GenerateUniqueCommandID: String;
var
    i: Integer;
    NewID: Integer;
    Success: Boolean;
begin
    repeat
        Success := True;
        NewID := Random(High(Integer));
        for i := 0 to CommandList.Count - 1 do begin
            if CommandList.List[i].ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
    until Success
    AND (NewID <> 0);
    Result := IntToStr(NewID);
end;

function TConnectedClient.GenerateUniqueTransferID: Integer;
var
    i: Integer;
    NewID: Integer;
    Success: Boolean;
begin
    repeat
        Success := True;
        NewID := Random(High(Integer));
        for i := 0 to DownloadList.Count - 1 do begin
            if TMessageBinary(DownloadList.List[i]).ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
        for i := 0 to SendList.Count - 1 do begin
            if TMessageBinary(SendList.List[i]).ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
        for i := 0 to SendListSynchronous.Count - 1 do begin
            if TMessageBinary(SendListSynchronous.List[i]).ID = NewID then begin
                Success := False;
                Break;
            end;
        end;
    until Success
    AND (NewID <> 0);
    Result := NewID;
end;

procedure TConnectedClient.LockTransferLists;
begin
    DownloadList.Lock;
    SendList.Lock;
end;

procedure TConnectedClient.UnLockTransferLists;
begin
    DownloadList.UnLock;
    SendList.UnLock;
end;

{ TMessageBinaryFile }

constructor TMessageBinaryFile.Create(FileName: String; _Resume: Boolean = False);
begin
    inherited Create;
    if FileExists(FileName) then begin
        if _Resume then begin
            Data := TFileStream.Create(FileName, fmOpenReadWrite OR fmShareDenyNone);
        end else begin
            Data := TFileStream.Create(FileName, fmOpenRead OR fmShareDenyNone);
        end
    end else begin
        Data := TFileStream.Create(FileName, fmCreate OR fmShareDenyWrite);
    end;
    LocalFileName := FileName;
    Resume := _Resume;
end;

{ TMessageBinaryMemory }

constructor TMessageBinaryMemory.Create;
begin
    inherited;
    Data := TMemoryStream.Create;
end;

constructor TMessageBinaryMemory.Create(Text: String);
begin
    inherited Create;
    {$IFDEF NEXTGEN}
    Data.Write(Text[0], Length(Text) * SizeOf(Text[0]));
    {$ELSE}
    Data.Write(Text[1], Length(Text) * SizeOf(Text[1]));
    {$ENDIF}
    Data.Seek(0, soBeginning);
end;

constructor TMessageBinaryMemory.Create(Strings: TStrings);
begin
    inherited Create;
    Strings.SaveToStream(Data);
    Data.Seek(0, soBeginning);
end;

constructor TMessageBinaryMemory.Create(Stream: TStream);
begin
    inherited Create;
    Stream.Seek(0, soBeginning);
    Data.CopyFrom(Stream, Stream.Size);
    Data.Seek(0, soBeginning);
end;

{ TClientSendList }

function TDownloadList.Add(Item: TMessageBinary): Integer;
begin
    Result := List.Add(Item);
end;

procedure TDownloadList.Clear;
var
    i: Integer;
begin
    for i := List.Count - 1 downto 0 do begin
        {$IFNDEF NEXTGEN}
        TObject(List[i]).Free;
        {$ENDIF}
        List.Delete(i);
    end;
end;

function TDownloadList.Count: Integer;
begin
    Result := List.Count;
end;

constructor TDownloadList.Create;
begin
    inherited;
    FCriticalSection := TCriticalSection.Create;
    {$IFDEF TCPSCL_GENERICS}
    List := TList<TMessageBinary>.Create;
    {$ELSE}
    List := TList.Create;
    {$ENDIF}
end;

procedure TDownloadList.Delete(Index: Integer);
begin
    List.Delete(Index);
end;

destructor TDownloadList.Destroy;
begin
    Clear;
    FreeAndNil(List);
    FreeAndNil(FCriticalSection);
    inherited;
end;

function TDownloadList.Find(ID: Integer): Integer;
var
    i: Integer;
begin
    Result := - 1;
    for i := 0 to List.Count - 1 do begin
        if TTCPServerMessageID(List[i]).ID = ID then begin
            Result := i;
            Break;
        end;
    end;
end;

procedure TDownloadList.Lock;
begin
    FCriticalSection.Enter;
end;

procedure TDownloadList.UnLock;
begin
    FCriticalSection.Leave;
end;

{ TCommandList }

procedure TCommandList.Add(Item: TMessageCommand){: TMessageCommand};
begin
    List.Add(Item);
    //Result := Item;
end;

procedure TCommandList.Clear;
var
    i: Integer;
begin
    for i := List.Count - 1 downto 0 do begin
        {$IFNDEF NEXTGEN}
        TObject(List[i]).Free;
        {$ENDIF}
        List.Delete(i);
    end;
end;

function TCommandList.Count: Integer;
begin
    Result := List.Count;
end;

constructor TCommandList.Create;
begin
    inherited;
    FCriticalSection := TCriticalSection.Create;
    {$IFDEF TCPSCL_GENERICS}
    List := TList<TMessageCommand>.Create;
    {$ELSE}
    List := TList.Create;
    {$ENDIF}
end;

procedure TCommandList.Delete(Index: Integer);
begin
    List.Delete(Index);
end;

destructor TCommandList.Destroy;
begin
    Clear;
    FreeAndNil(List);
    FreeAndNil(FCriticalSection);
    inherited;
end;

function TCommandList.Find(ID: Integer): Integer;
var
    i: Integer;
begin
    Result := - 1;
    for i := 0 to List.Count - 1 do begin
        if List[i].ID = ID then begin
            Result := i;
            Break;
        end;
    end;
end;

procedure TCommandList.Lock;
begin
    FCriticalSection.Enter;
end;

procedure TCommandList.UnLock;
begin
    FCriticalSection.Leave;
end;

{ TMessageCommand }

destructor TMessageCommand.Destroy;
begin
    //if Assigned(Data) then begin
    //    FreeAndNil(Data);
    //end;
    inherited;
end;

{ TTCPServerParams }

constructor TTCPServerParams.Create;
begin
    inherited;
    FTransferDataInvalidInterval := TRANSFER_DATA_INVALID_INTERVAL;
    FClientTimeoutDelay := CLIENT_TIMEOUT_DELAY;
    FDataTransferPacketSize := DATA_TRANSFER_PACKET_SIZE;
    FClientPingInterval := CLIENT_PING_INTERVAL;
    FTransferDataPurgeInterval := TRANSFER_DATA_PURGE_INTERVAL;
end;

{ TTCPClientParams }

constructor TTCPClientParams.Create;
begin
    inherited;
    FTransferDataInvalidInterval := TRANSFER_DATA_INVALID_INTERVAL;
    FDataTransferPacketSize := DATA_TRANSFER_PACKET_SIZE;
    FTransferDataPurgeInterval := TRANSFER_DATA_PURGE_INTERVAL;
end;

Initialization

    Randomize;

end.

