//********************************************************************************************************************************
//*                                                                                                                              *
//*     TCP Server-Client Library 1.0.40.80  3delite 2013-2025                                                                  *
//*     See TCP Server-Client Library ReadMe.txt for details.                                                                    *
//*                                                                                                                              *
//* Licenses available for usage of this component:                                                                              *
//* Freeware License: 25                                                                                                        *
//* Shareware License: 50                                                                                                       *
//* Commercial License: 250                                                                                                     *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/register.html#tcpscl                                      *
//*                                                                                                                              *
//* Home page:                                                                                                                   *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/TCPServerClientLibrary.html                               *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     https://www.3delite.hu/                                                                                                  *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit TCPServerThread;

interface

{$I TCPSCLCompilerDefines.inc}

uses
    SysUtils,
    Classes,
    SyncObjs,
    {$IFDEF TCPSCL_MOBILE}
    System.Generics.Collections,
    {$ENDIF}
    ThreadMessageQueue,
    IdGlobal,
    IdTCPServer,
    IdContext,
    IdSocketHandle,
    IdServerIOHandler,
    IdSSLOpenSSL,
    IdSSL,
    HashFunctions,
    StrUtils,
    DateUtils,
    TCPLibraryClasses,
    UsersList;

type
    TTCPServerParameters = record
        Parent: TObject;
        ServerName: String;
        DefaultPort: Cardinal;
        MaxConnections: Cardinal;
        LoggingOn: Boolean;
        SSLTLSUse: Boolean;
        SSLPassword: String;
        SSLRootCertFile: String;
        SSLCertFile: String;
        SSLKeyFile: String;
        LoginHashingMode: TTCPSCLHashingMode;
        SendUserListToClients: Boolean;
        LoginList: TLoginList;
        Bindings: TIdSocketHandles;
        Parameters: TTCPServerParams;
        ActionMessageQueue: TThreadMessageQueue;
        IncomingMessageQueue: TThreadMessageQueue;
        OnServerCommandReceive: TServerCommandReceiveEvent;
        OnServerClientConnected: TServerClientConnectedEvent;
        OnServerClientDisconnected: TServerClientDisconnectedEvent;
        OnServerLoginQuery: TServerLoginQueryEvent;
        OnServerBeforeActivate: TServerBeforeActivateEvent;
        OnServerAfterActivate: TServerAfterActivateEvent;
        OnServerBeforeDeActivate: TServerBeforeDeActivateEvent;
        OnServerAfterDeActivate: TServerAfterDeActivateEvent;
        OnServerThreadStart: TServerThreadStartEvent;
        OnServerThreadEnd: TServerThreadEndEvent;
        OnServerCustomSend: TServerCustomSendEvent;
    end;

type
    TTCPServerThread = class(TThread)
    private
        FServerName: String;
        FSSLPassword: String;
        FLoginHashingMode: TTCPSCLHashingMode;
        FSSLIOHandler: TIdServerIOHandlerSSLOpenSSL;
        FTimeOutCleanUpLastDateTime: TDateTime;
        FOnServerCommandReceive: TServerCommandReceiveEvent;
        FOnServerClientConnected: TServerClientConnectedEvent;
        FOnServerClientDisconnected: TServerClientDisconnectedEvent;
        FOnServerLoginQuery: TServerLoginQueryEvent;
        FOnServerBeforeActivate: TServerBeforeActivateEvent;
        FOnServerAfterActivate: TServerAfterActivateEvent;
        FOnServerBeforeDeActivate: TServerBeforeDeActivateEvent;
        FOnServerAfterDeActivate: TServerAfterDeActivateEvent;
        FOnServerThreadStart: TServerThreadStartEvent;
        FOnServerThreadEnd: TServerThreadEndEvent;
        FOnServerCustomSend: TServerCustomSendEvent;
        FParameters: TTCPServerParams;
        procedure IdTCPServerConnect(AContext: TIdContext);
        procedure IdTCPServerDisconnect(AContext: TIdContext);
        procedure IdTCPServerExecute(AContext: TIdContext);
        {$IFDEF TCPSCL_DELPHI_XE5 OR TCPSCL_MOBILE}
        procedure SSLGetPassword(var Password: String);
        {$ELSE}
        procedure SSLGetPassword(var Password: AnsiString);
        {$ENDIF}
        procedure Log(const Text: String; Debug: Boolean);
        procedure SendBinary(MessageBinary: TMessageBinary);
        procedure SendCommand(MessageCommand: TMessageCommand);
        procedure SendChat(MessageChat: TMessageChat);
        procedure SendClientList;
        procedure LoopbackChat(MessageChat: TMessageChat);
        procedure SendTerminated;
        procedure TransferCancel(MessageTransferCancel: TMessageTransferCancel);
        procedure ProcessTransferRequest(MessageTransferRequest: TMessageTransferRequest);
        procedure ClientKick(MessageClientKick: TMessageClientKick);
        procedure ReportProgress(Context: TIdContext; Progress: Integer; TransferID: Integer; ElapsedTime, EstimatedTime: TDateTime; TransferRate: Cardinal);
        procedure DeleteTimeOutPackets;
        function DataTransferAgreed(BinaryPacketID: Integer; List: TSendList): Boolean;
        function DataTransferCancel(BinaryPacketID: Integer; List: TSendList): Boolean;
        function DoBinaryTransfer(AContext: TIdContext; MessageBinary: TMessageBinary; List: TSendList): Boolean;
    public
        Parent: TObject;
        IdTCPServer: TIdTCPServer;
        ConnectedClients: TConnectedClients;
        {$IFDEF INDY_ENCODING}
        ByteEncoding: IIdTextEncoding;
        {$ELSE}
        ByteEncoding: TEncoding;
        {$ENDIF}
        LoginList: TLoginList;
        LoggingOn: Boolean;
        ActionMessageQueue: TThreadMessageQueue;
        IncomingMessageQueue: TThreadMessageQueue;
        SendUserListToClients: Boolean;
        Constructor Create(TCPServerParameters: TTCPServerParameters);
        Destructor Destroy; override;
        function Active: Boolean;
    protected
        procedure Execute; override;
    end;

implementation

Uses
    IdStack;

constructor TTCPServerThread.Create(TCPServerParameters: TTCPServerParameters);
begin
    inherited Create(False);
    FreeOnTerminate := False;
    {$IFDEF INDY_ENCODING}
        {$IFDEF INDY_ENCODING_XE5}
        ByteEncoding := IndyTextEncoding_UTF8;
        {$ELSE}
        ByteEncoding := IndyUTF8Encoding;
        {$ENDIF}
    {$ELSE}
    ByteEncoding := TUTF8Encoding.Create;
    {$ENDIF}
    FServerName := TCPServerParameters.ServerName;
    ConnectedClients := TConnectedClients.Create;
    Parent := TCPServerParameters.Parent;
    LoggingOn := TCPServerParameters.LoggingOn;
    FParameters := TCPServerParameters.Parameters;
    //* Create Indy TCP server object
    IdTCPServer := TIdTCPServer.Create;
    IdTCPServer.DefaultPort := TCPServerParameters.DefaultPort;
    IdTCPServer.Bindings.Assign(TCPServerParameters.Bindings);
    //* If no binding is specified use the (default) first IP address (IPv4)
    if IdTCPServer.Bindings.Count = 0 then begin
        TIdStack.IncUsage;
        try
            with IdTCPServer.Bindings.Add do begin
                IPVersion := Id_IPv4;
                IP := GStack.LocalAddress;
                Port := IdTCPServer.DefaultPort;
            end;
        finally
            TIdStack.DecUsage;
        end;
    end;
    IdTCPServer.MaxConnections := TCPServerParameters.MaxConnections;
    IdTCPServer.OnConnect := IdTCPServerConnect;
    IdTCPServer.OnDisconnect := IdTCPServerDisconnect;
    IdTCPServer.OnExecute := IdTCPServerExecute;
    LoginList := TCPServerParameters.LoginList;
    ActionMessageQueue := TCPServerParameters.ActionMessageQueue;
    IncomingMessageQueue := TCPServerParameters.IncomingMessageQueue;
    FOnServerCommandReceive := TCPServerParameters.OnServerCommandReceive;
    FOnServerClientConnected := TCPServerParameters.OnServerClientConnected;
    FOnServerClientDisconnected := TCPServerParameters.OnServerClientDisconnected;
    FOnServerLoginQuery := TCPServerParameters.OnServerLoginQuery;
    FOnServerBeforeActivate := TCPServerParameters.OnServerBeforeActivate;
    FOnServerAfterActivate := TCPServerParameters.OnServerAfterActivate;
    FOnServerBeforeDeActivate := TCPServerParameters.OnServerBeforeDeActivate;
    FOnServerAfterDeActivate := TCPServerParameters.OnServerAfterDeActivate;
    FOnServerThreadStart := TCPServerParameters.OnServerThreadStart;
    FOnServerThreadEnd := TCPServerParameters.OnServerThreadEnd;
    FOnServerCustomSend := TCPServerParameters.OnServerCustomSend;
    //* SSL
    if TCPServerParameters.SSLTLSUse then begin
        FSSLPassword := TCPServerParameters.SSLPassword;
        FSSLIOHandler := TIdServerIOHandlerSSLOpenSSL.Create(nil);
        with FSSLIOHandler do begin
            SSLOptions.SSLVersions := [sslvTLSv1]; //* default
            SSLOptions.RootCertFile := TCPServerParameters.SSLRootCertFile;
            SSLOptions.CertFile := TCPServerParameters.SSLCertFile;
            SSLOptions.KeyFile := TCPServerParameters.SSLKeyFile;
            OnGetPassword := SSLGetPassword;
        end;
        IdTCPServer.IOHandler := FSSLIOHandler;
    end;
    FLoginHashingMode := TCPServerParameters.LoginHashingMode;
    SendUserListToClients := TCPServerParameters.SendUserListToClients;
end;

procedure TTCPServerThread.DeleteTimeOutPackets;
var
    i, k: Integer;
    CurrentDateTime: TDateTime;
    {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
    ContextList: System.Generics.Collections.TList<TIdContext>;
    {$ELSE}
    ContextList: TList;
    {$ENDIF}
    Context: TIdContext;
    ConnectedClient: TConnectedClient;
begin
    CurrentDateTime := Now;
    ContextList := IdTCPServer.Contexts.LockList;
    try
        for i := 0 to ContextList.Count - 1 do begin
            Context := TObject(ContextList[i]) as TIdContext;
            ConnectedClients.Lock;
            try
                ConnectedClient := ConnectedClients.Find(Context);
                if Assigned(ConnectedClient) then begin
                    //* Asynchronous
                    ConnectedClient.SendList.Lock;
                    try
                        for k := ConnectedClient.SendList.Count - 1 downto 0 do begin
                            if MilliSecondsBetween(CurrentDateTime, TMessageBinary(ConnectedClient.SendList.List[k]).LastTouchTime) > FParameters.TransferDataInvalidInterval then begin
                                {$IFNDEF NEXTGEN}
                                TMessageBinary(ConnectedClient.SendList.List[k]).Free;
                                {$ENDIF}
                                ConnectedClient.SendList.Delete(k);
                            end;
                        end;
                    finally
                        ConnectedClient.SendList.UnLock;
                    end;
                    //* Synchronous
                    ConnectedClient.SendListSynchronous.Lock;
                    try
                        for k := ConnectedClient.SendListSynchronous.Count - 1 downto 0 do begin
                            if MilliSecondsBetween(CurrentDateTime, TMessageBinary(ConnectedClient.SendListSynchronous.List[k]).LastTouchTime) > FParameters.TransferDataInvalidInterval then begin
                                {$IFNDEF NEXTGEN}
                                TMessageBinary(ConnectedClient.SendListSynchronous.List[k]).Free;
                                {$ENDIF}
                                ConnectedClient.SendListSynchronous.Delete(k);
                            end;
                        end;
                    finally
                        ConnectedClient.SendListSynchronous.UnLock;
                    end;
                end;
            finally
                ConnectedClients.UnLock;
            end;
        end;
    finally
        IdTCPServer.Contexts.UnlockList;
    end;
end;

destructor TTCPServerThread.Destroy;
begin
    //IdTCPServer.Active := False;
    FreeAndNil(ConnectedClients);
    FreeAndNil(IdTCPServer);
    if Assigned(FSSLIOHandler) then begin
        FreeAndNil(FSSLIOHandler);
    end;
    {$IFDEF INDY_ENCODING}
    //FreeAndNil(ByteEncoding);
    {$ELSE}
    FreeAndNil(ByteEncoding);
    {$ENDIF}
    inherited;
end;

procedure TTCPServerThread.Execute;
var
    i: Integer;
    Action: TObject;
begin
    try
        if Assigned(FOnServerThreadStart) then begin
            FOnServerThreadStart(Self.Parent);
        end;
        Log(FServerName + ' starting...', True);
        if Assigned(FOnServerBeforeActivate) then begin
            FOnServerBeforeActivate(Self.Parent);
        end;
        IdTCPServer.Active := True;
        if IdTCPServer.Active then begin
            if Assigned(FOnServerAfterActivate) then begin
                FOnServerAfterActivate(Self.Parent);
            end;
            Log(FServerName + ' started on the following address(es):', True);
            for i := 0 to IdTCPServer.Bindings.Count - 1 do begin
                Log(IdTCPServer.Bindings[i].IP + ':' + IntToStr(IdTCPServer.Bindings[i].Port) {+ ' ' + BoolToStr(IdTCPServer.Bindings[i].HandleAllocated, True)}, True);
            end;
        end else begin
            Log('Error while starting server ' + FServerName, True);
        end;
        FTimeOutCleanUpLastDateTime := Now;
        while (NOT Terminated)
        AND IdTCPServer.Active
        do begin
            //* Send message queue
            if ActionMessageQueue.Count > 0 then begin
                Action := ActionMessageQueue.Pop;
                if Action is TMessageChat then begin
                    SendChat(TMessageChat(Action));
                    LoopbackChat(TMessageChat(Action));
                    {$IFNDEF NEXTGEN}
                    FreeAndNil(Action);
                    {$ENDIF}
                end else if Action is TMessageCommand then begin
                    SendCommand(TMessageCommand(Action));
                end else if Action is TMessageBinary then begin
                    SendBinary(TMessageBinary(Action));
                end else if Action is TMessageTransferRequest then begin
                    ProcessTransferRequest(TMessageTransferRequest(Action));
                end else if Action is TMessageTransferCancel then begin
                    TransferCancel(TMessageTransferCancel(Action));
                end else if Action is TMessageClientKick then begin
                    ClientKick(TMessageClientKick(Action));
                end;
            end else begin
                Sleep(15);
            end;
            //* Message data clean up cycle
            if MilliSecondsBetween(Now, FTimeOutCleanUpLastDateTime) > FParameters.TransferDataPurgeInterval then begin
                DeleteTimeOutPackets;
                FTimeOutCleanUpLastDateTime := Now;
            end;
        end;
        if IdTCPServer.Active then begin
            if Assigned(FOnServerBeforeDeActivate) then begin
                FOnServerBeforeDeActivate(Self.Parent);
            end;
            IdTCPServer.Active := False;
            if Assigned(FOnServerAfterDeActivate) then begin
                FOnServerAfterDeActivate(Self.Parent);
            end;
        end;
    except
        On E: exception do begin
            Log('Exception: ' + E.Message, True);
        end;
    end;
    if Assigned(FOnServerThreadEnd) then begin
        FOnServerThreadEnd(Self.Parent);
    end;
    SendTerminated;
    Log(FServerName + ' terminated.', True);
end;

procedure TTCPServerThread.IdTCPServerConnect(AContext: TIdContext);
var
    Salt: String;
    PassWord: String;
    NeededPassWord: String;
    ReceivedPassWord: String;
    LoginName: String;
    UserName: String;
    LoginRights: UInt64;
    Login: TLogin;
    ConnectionInfo: TConnectionInfo;
    ConnectedClient: TConnectedClient;
    Deny: Boolean;
    i: Integer;
begin
    Log('A connection has been made by a client: ' + AContext.Binding.PeerIP + ':' + IntToStr(AContext.Binding.PeerPort), True);

    { THESE TWO LINES ARE CRITICAL TO MAKING THE IdTCPSERVER WORK WITH SSL! }
    if Assigned(FSSLIOHandler) then begin
        if (AContext.Connection.IOHandler is TIdSSLIOHandlerSocketBase) then begin
            TIdSSLIOHandlerSocketBase(AContext.Connection.IOHandler).PassThrough := false;
        end;
    end;

    AContext.Connection.Socket.ReadTimeout := FParameters.ClientTimeoutDelay;
    AContext.Connection.Socket.SendBufferSize := FParameters.DataTransferPacketSize;
    //AContext.Connection.Socket.UseNagle := False;
    LoginRights := 0;
    Login := nil;
    Deny := False;
    try
        AContext.Connection.IOHandler.WriteLn(FServerName, ByteEncoding);
        AContext.Connection.IOHandler.WriteLn('511 Login name?', ByteEncoding);
        LoginName := AContext.Connection.IOHandler.ReadLn(ByteEncoding);
        LoginName := Trim(LoginName);
        try
            //* Find user
            if Assigned(FOnServerLoginQuery) then begin
                FOnServerLoginQuery(Parent, LoginName, PassWord, Deny);
            end else begin
                LoginList.Lock;
                try
                    Login := nil;
                    if LoginList.Find(LoginName, Login) then begin
                        PassWord := Login.PassWord;
                        LoginRights := Login.Rights;
                    end else begin
                        Deny := True;
                    end;
                finally
                    LoginList.UnLock;
                end;
            end;
            Salt := FloatToStr(Now);
            for i := 1 to 10 do begin
                Salt := Salt + FloatToStr(Random);
            end;
            case FLoginHashingMode of
                hmOpenSSL: begin
                    Salt := HashSHA512String(Salt);
                    NeededPassWord := HashSHA512String(Salt + PassWord);
                end;
                hmMD5: begin
                    Salt := HashMD5String(Salt);
                    NeededPassWord := HashMD5String(Salt + PassWord);
                end;
            end;
            AContext.Connection.IOHandler.WriteLn(Salt, ByteEncoding);
            ReceivedPassWord := AContext.Connection.IOHandler.ReadLn(ByteEncoding);
            if (Trim(ReceivedPassWord) = NeededPassWord)
            AND (NOT Deny)
            then begin
                AContext.Connection.IOHandler.WriteLn('200 OK', ByteEncoding);
                Log(AContext.Binding.PeerIP + ':' + IntToStr(AContext.Binding.PeerPort) + ' client confirmed, login: ' + LoginName, True);
                AContext.Connection.IOHandler.WriteLn('100 User name?', ByteEncoding);
                UserName := AContext.Connection.IOHandler.ReadLn(ByteEncoding);
                UserName := Trim(UserName);
                ConnectedClients.Lock;
                ConnectedClients.Add(AContext, LoginName, UserName, 0, LoginRights);
                ConnectedClients.UnLock;
                ConnectionInfo := TConnectionInfo.Create;
                ConnectionInfo.LastHeartBeatDateTime := Now;
                {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
                AContext.DataObject := ConnectionInfo;
                {$ELSE}
                AContext.Data := ConnectionInfo;
                {$ENDIF}
                AContext.Connection.IOHandler.WriteLn('200 OK', ByteEncoding);
                Log(AContext.Binding.PeerIP + ':' + IntToStr(AContext.Binding.PeerPort) + ' client user name is: ' + UserName, True);
                if Assigned(FOnServerClientConnected) then begin
                    ConnectedClients.Lock;
                    try
                        ConnectedClient := ConnectedClients.Find(AContext);
                        FOnServerClientConnected(Parent, AContext, ConnectedClient, ByteEncoding);
                    finally
                        ConnectedClients.UnLock;
                    end;
                end;
                if SendUserListToClients then begin
                   SendClientList;
                end;
            end else begin
                AContext.Connection.IOHandler.WriteLn('401 Unknown/rejected login name and/or wrong password.', ByteEncoding);
                AContext.Connection.Disconnect;
            end;
        finally
            if Assigned(Login) then begin
                FreeAndNil(Login);
            end;
        end;
    except
        On E: exception do begin
            Log('Exception: ' + E.Message, True);
        end;
    end;
end;

procedure TTCPServerThread.IdTCPServerDisconnect(AContext: TIdContext);
var
    ConnectedClient: TConnectedClient;
begin
    ConnectedClients.Lock;
    try
        ConnectedClient := ConnectedClients.Find(AContext);
        if Assigned(FOnServerClientDisconnected) then begin
            FOnServerClientDisconnected(Parent, AContext, ConnectedClient);
        end;
        if Assigned(ConnectedClient) then begin
            Log(AContext.Binding.PeerIP + ' - ' + ConnectedClient.UserName + ' has left :(', True);
            ConnectedClients.Remove(AContext);
        end else begin
            Log('Unknown user ' + AContext.Binding.PeerIP + ' has left :(', True);
        end;
        if SendUserListToClients then begin
            SendClientList;
        end;
    finally
        ConnectedClients.UnLock;
    end;
end;

function TTCPServerThread.DataTransferAgreed(BinaryPacketID: Integer; List: TSendList): Boolean;
var
    BinaryPacketIndex: Integer;
    MessageBinary: TMessageBinary;
begin
    Result := False;
    List.Lock;
    try
        BinaryPacketIndex := List.Find(BinaryPacketID);
        if BinaryPacketIndex > - 1 then begin
            MessageBinary := TMessageBinary(List.List[BinaryPacketIndex]);
            MessageBinary.Data.Seek(MessageBinary.StartPosition, soBeginning);
            MessageBinary.DoTransfer := True;
            Result := True;
        end;
    finally
        List.UnLock;
    end;
end;

function TTCPServerThread.DataTransferCancel(BinaryPacketID: Integer; List: TSendList): Boolean;
var
    BinaryPacketIndex: Integer;
    MessageBinary: TMessageBinary;
begin
    Result := False;
    List.Lock;
    try
        BinaryPacketIndex := List.Find(BinaryPacketID);
        if BinaryPacketIndex > - 1 then begin
            MessageBinary := TMessageBinary(List.List[BinaryPacketIndex]);
            IncomingMessageQueue.Add(MessageBinary);
            Log('Binary transfer canceled: ' + IntToStr(BinaryPacketID) + ' ' + MessageBinary.FileName, True);
            List.Delete(BinaryPacketIndex);
            Result := True;
        end;
    finally
        List.UnLock;
    end;
end;

function TTCPServerThread.DoBinaryTransfer(AContext: TIdContext; MessageBinary: TMessageBinary; List: TSendList): Boolean;
var
    DataSizeToSend: Integer;
    EstimatedTime: TDateTime;
    ElapsedTime: TDateTime;
    TransferRate: Cardinal;
    Index: Integer;
    FirstPacket: Boolean;
begin
    Result := False;
    if MessageBinary.Arguments = '' then begin
        MessageBinary.Arguments := '0';
    end;
    //* Send a send request
    if (NOT MessageBinary.RequestSent)
    AND (MessageBinary.RequireAgreement = raTrue)
    then begin
        AContext.Connection.IOHandler.WriteLn(IntToStr(MessageBinary.ID) + ' 800 ' + IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size) + ' ' + IntToStr(MessageBinary.StartPosition) + ' "' + MessageBinary.FileName + '"' + ' ' + MessageBinary.Arguments, ByteEncoding);
        MessageBinary.RequestSent := True;
        Result := True;
        Exit;
    end;
    //* Send the next data packet
    if MessageBinary.DoTransfer then begin
        with AContext.Connection.IOHandler do begin
            FirstPacket := False;
            if MessageBinary.DownloadStartTime = 0 then begin
                MessageBinary.DownloadStartTime := Now;
                MessageBinary.DownloadProgressMax := MessageBinary.TotalSize - MessageBinary.StartPosition;
                MessageBinary.DownloadLastReport := Now;
                ReportProgress(AContext, 0, MessageBinary.ID, 0, 0, 0);
                Log('Upload started: ' + MessageBinary.FileName, True);
                FirstPacket := True;
            end;
            if MessageBinary.Data.Size - MessageBinary.Data.Position > FParameters.DataTransferPacketSize then begin
                DataSizeToSend := FParameters.DataTransferPacketSize;
            end else begin
                DataSizeToSend := MessageBinary.Data.Size - MessageBinary.Data.Position;
            end;
            if FirstPacket then begin
                WriteLn(IntToStr(MessageBinary.ID) + ' 802 ' + IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size) + ' "' + MessageBinary.FileName + '"' + ' ' + MessageBinary.Arguments, ByteEncoding);
            end else begin
                WriteLn(IntToStr(MessageBinary.ID) + ' 802' {+ IntToStr(MessageBinary.DataType) + ' ' + IntToStr(MessageBinary.Data.Size)}, ByteEncoding);
            end;
            WriteLn(IntToStr(DataSizeToSend), ByteEncoding);
            Write(MessageBinary.Data, DataSizeToSend);
            with MessageBinary do begin
                LastTouchTime := Now;
                //* Report every 250ms
                if DateTimeDiffmSec(DownloadLastReport, Now) > 250 then begin
                    DownloadProgress := Round(((Data.Position - MessageBinary.StartPosition) / DownloadProgressMax) * 100);
                    EstimatedTime := ((Now - DownloadStartTime) * DownloadProgressMax) / (Data.Position - MessageBinary.StartPosition);
                    ElapsedTime := Now - DownloadStartTime;
                    if ElapsedTime <> 0 then begin
                        TransferRate := Round((MessageBinary.Data.Position  - MessageBinary.StartPosition )/ SecondSpan(ElapsedTime, 0));
                    end else begin
                        TransferRate := MessageBinary.Data.Position - MessageBinary.StartPosition;
                    end;
                    ReportProgress(AContext, DownloadProgress, ID, ElapsedTime, EstimatedTime, TransferRate);
                    DownloadLastReport := Now;
                end;
            end;
            Result := True;
        end;
        //* Completed
        if MessageBinary.Data.Position = MessageBinary.Data.Size then begin
            MessageBinary.DownloadProgress := 100;
            ElapsedTime := Now - MessageBinary.DownloadStartTime;
            if ElapsedTime <> 0 then begin
                TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition) / SecondSpan(ElapsedTime, 0));
            end else begin
                TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
            end;
            ReportProgress(AContext, 100, MessageBinary.ID, ElapsedTime, 0, TransferRate);
            Log('Successfully sent binary: ' + MessageBinary.FileName, True);
            Index := List.Find(MessageBinary.ID);
            if (Index > - 1)
            AND (Index < List.Count)
            then begin
                List.Delete(Index);
            end;
            {$IFNDEF NEXTGEN}
            FreeAndNil(MessageBinary);
            {$ENDIF}
        end;
    end;
end;

procedure TTCPServerThread.IdTCPServerExecute(AContext: TIdContext);
var
    ClientData: String;
    ConnectedClient: TConnectedClient;
    //NewProtocolLevel: Cardinal;
    BinaryPacketID: Integer;
    //BinaryPacketIndex: Integer;
    Command: TStringList;
    i: Integer;
    //ClientMessageBinary: TMessageBinary;
    //DataSizeToSend: Integer;
    MessageChat: TMessageChat;
    MessageChatLocal: TMessageChat;
    MessageCommand: TMessageCommand;
    Handled: Boolean;
    MessageTransferRequest: TMessageTransferRequest;
    MessageBinaryID: Integer;
    Size: Integer;
    MessageBinaryIndex: Integer;
    MessageBinary: TMessageBinary;
    EstimatedTime: TDateTime;
    ElapsedTime: TDateTime;
    MessageTransferCancel: TMessageTransferCancel;
    TransferRate: Cardinal;
    DoReceiveBinary: Boolean;
    TempReceiveStream: TMemoryStream;
begin
    Handled := False;
    MessageBinary := nil;
    Command := TStringList.Create;
    Command.Delimiter := ' ';
    ConnectedClient := ConnectedClients.Find(AContext);

    try
        try
            //* Send custom command
            if Assigned(FOnServerCustomSend) then begin
                FOnServerCustomSend(Parent, AContext, ConnectedClient, ByteEncoding);
            end;

            //* If there is incomming data, process it
            ClientData := AContext.Connection.IOHandler.ReadLn('', 1, - 1, ByteEncoding);
            ClientData := Trim(ClientData);
            if ClientData <> '' then begin

                //* Send the command to the component
                if Assigned(FOnServerCommandReceive) then begin
                    FOnServerCommandReceive(Parent, AContext, ConnectedClient, ClientData, ByteEncoding, Handled);
                end;

                if NOT Handled then begin

                    //* Keepalive
                    if SameText(ClientData, 'PONG') then begin
                        //* Nothing
                        Exit;
                    end;

                    //* "Decode" the command
                    Command.DelimitedText := ClientData;

                    //* Very bad, should not happen
                    //if Command.Count <= 1 then begin
                    //    AContext.Connection.Disconnect;
                    //    Exit;
                    //end;

                    //* Client wants to set protocol level
                    {
                    if SameText(Command[1], '999')
                    then begin
                        NewProtocolLevel := StrToIntDef(Command[2], 0);
                        if (NewProtocolLevel > 0)
                        AND (NewProtocolLevel <= MAX_PROTOCOL_LEVEL)
                        then begin
                            ConnectedClient.ProtocolLevel := NewProtocolLevel;
                            AContext.Connection.IOHandler.WriteLn(Command[0] + ' 200 OK', ByteEncoding);
                            Log(ConnectedClient.UserName + ' set new protocol level: ' + IntToStr(NewProtocolLevel), True);
                        end else begin
                            AContext.Connection.IOHandler.WriteLn(Command[0] + ' 501', ByteEncoding);
                        end;
                        AContext.Connection.IOHandler.WriteLn(Command[0] + ' 101 ' + IntToStr(ConnectedClient.ProtocolLevel), ByteEncoding);
                        Exit;
                    end;
                    }

                    //* Chat
                    if SameText(Command[1], '111')
                    then begin
                        MessageChat := TMessageChat.Create;
                        MessageChat.UserName := ConnectedClient.UserName;
                        MessageChat.Text := Copy(ClientData, 7, Length(ClientData));
                        SendChat(MessageChat);
                        FreeAndNil(MessageChat);
                        MessageChatLocal := TMessageChat.Create;
                        MessageChatLocal.UserName := ConnectedClient.UserName;
                        MessageChatLocal.Text := Copy(ClientData, 7, Length(ClientData));
                        IncomingMessageQueue.Add(MessageChatLocal);
                        Exit;
                    end;

                    //* Data transfer cancel
                    if SameText(Command[1], '809') then begin
                        BinaryPacketID := StrToIntDef(Command[0], 0);
                        if NOT DataTransferCancel(BinaryPacketID, ConnectedClient.SendList) then begin
                            DataTransferCancel(BinaryPacketID, ConnectedClient.SendListSynchronous);
                        end;
                        Exit;
                    end;

                    //* Very bad, should not happen
                    //if Command.Count <= 2 then begin
                    //    AContext.Connection.Disconnect;
                    //    Exit;
                    //end;

                    //* Data transfer agreed
                    if SameText(Command[1], '200')
                    AND SameText(Command[2], '801')
                    then begin
                        BinaryPacketID := StrToIntDef(Command[0], 0);
                        if NOT DataTransferAgreed(BinaryPacketID, ConnectedClient.SendList) then begin
                            DataTransferAgreed(BinaryPacketID, ConnectedClient.SendListSynchronous);
                        end;
                        Exit;
                    end;

                    //* Transfer request arrived
                    if SameText(Command[1], '800') then begin
                        MessageTransferRequest := TMessageTransferRequest.Create;
                        MessageTransferRequest.DataType := StrToIntDef(Command[2], 0);
                        MessageTransferRequest.ID := StrToIntDef(Command[0], 0);
                        MessageTransferRequest.TotalSize := StrToIntDef(Command[3], 0);
                        MessageTransferRequest.StartPosition := StrToIntDef(Command[4], 0);
                        MessageTransferRequest.FileName := Copy(ClientData, Pos('"', ClientData) + 1, PosEx('"', ClientData, Pos('"', ClientData) + 1) - Pos('"', ClientData) - 1);
                        MessageTransferRequest.Arguments := Copy(ClientData, PosEx('"', ClientData, Pos('"', ClientData) + 1) + 2, Length(ClientData));
                        MessageTransferRequest.Context := AContext;
                        IncomingMessageQueue.Add(MessageTransferRequest);
                        Exit;
                    end;

                    //* Transfer is in progress
                    if SameText(Command[1], '802') then begin
                        MessageBinaryID := StrToIntDef(Command[0], - 1);
                        Size := StrToInt64(Trim(AContext.Connection.IOHandler.ReadLn(ByteEncoding)));
                        ConnectedClient.DownloadList.Lock;
                        try
                            MessageBinaryIndex := ConnectedClient.DownloadList.Find(MessageBinaryID);
                            //* Asynchronous and/or already started synchronous
                            if MessageBinaryIndex > - 1 then begin
                                MessageBinary := TMessageBinary(ConnectedClient.DownloadList.List[MessageBinaryIndex]);
                                DoReceiveBinary := True;
                            //* New synchronous without agreement
                            end else begin
                                if Command.Count > 2 then begin
                                    MessageBinary := TMessageBinaryMemory.Create;
                                    MessageBinary.ID := MessageBinaryID;
                                    MessageBinary.DataType := StrToIntDef(Command[2], - 1);
                                    MessageBinary.TotalSize := StrToIntDef(Command[3], 0);
                                    //MessageBinary.StartPosition := StrToIntDef(Command[4], 0);
                                    MessageBinary.FileName := Copy(ClientData, Pos('"', ClientData) + 1, PosEx('"', ClientData, Pos('"', ClientData) + 1) - Pos('"', ClientData) - 1);
                                    MessageBinary.Arguments := Copy(ClientData, PosEx('"', ClientData, Pos('"', ClientData) + 1) + 2, Length(ClientData));
                                    MessageBinaryIndex := ConnectedClient.DownloadList.Add(MessageBinary);
                                    DoReceiveBinary := True;
                                end else begin
                                    DoReceiveBinary := False;
                                end;
                            end;
                            if DoReceiveBinary then begin
                                if MessageBinary.DownloadStartTime = 0 then begin
                                    MessageBinary.DownloadStartTime := Now;
                                    MessageBinary.DownloadProgressMax := MessageBinary.TotalSize - MessageBinary.StartPosition;
                                    MessageBinary.DownloadLastReport := Now;
                                    ReportProgress(AContext, 0, MessageBinaryID, 0, 0, 0);
                                    Log('Download started: ' + MessageBinary.FileName, True);
                                end;
                                try
                                    TempReceiveStream := TMemoryStream.Create;
                                    try
                                        AContext.Connection.IOHandler.ReadStream(TempReceiveStream, Size);
                                        MessageBinary.Data.CopyFrom(TempReceiveStream, 0);
                                    finally
                                        FreeAndNil(TempReceiveStream);
                                    end;
                                except
                                    On E: Exception do begin
                                        Log('Exception while receiving binary stream: ' + E.Message, True);
                                    end;
                                end;
                                with MessageBinary do begin
                                    LastTouchTime := Now;
                                    //* Report every 250ms
                                    if DateTimeDiffmSec(DownloadLastReport, Now) > 250 then begin
                                        DownloadProgress := Round(((MessageBinary.Data.Size - MessageBinary.StartPosition) / DownloadProgressMax) * 100);
                                        EstimatedTime := ((Now - DownloadStartTime) * DownloadProgressMax) / (MessageBinary.Data.Size - MessageBinary.StartPosition);
                                        ElapsedTime := Now - DownloadStartTime;
                                        if ElapsedTime <> 0 then begin
                                            TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition) / SecondSpan(ElapsedTime, 0));
                                        end else begin
                                            TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
                                        end;
                                        ReportProgress(AContext, DownloadProgress, MessageBinaryID, ElapsedTime, EstimatedTime, TransferRate);
                                        DownloadLastReport := Now;
                                    end;
                                end;
                                if MessageBinary.Data.Size = MessageBinary.TotalSize then begin
                                    ConnectedClient.DownloadList.Delete(MessageBinaryIndex);
                                    MessageBinary.DownloadProgress := 100;
                                    ElapsedTime := Now - MessageBinary.DownloadStartTime;
                                    if ElapsedTime <> 0 then begin
                                        TransferRate := Round((MessageBinary.Data.Size - MessageBinary.StartPosition) / SecondSpan(ElapsedTime, 0));
                                    end else begin
                                        TransferRate := MessageBinary.Data.Size - MessageBinary.StartPosition;
                                    end;
                                    MessageBinary.Data.Seek(0, soBeginning);
                                    ReportProgress(AContext, 100, MessageBinaryID, ElapsedTime, 0, TransferRate);
                                    Log('Download completed: ' + MessageBinary.FileName, True);
                                    IncomingMessageQueue.Add(MessageBinary);
                                end;
                            end else begin
                                AContext.Connection.IOHandler.Discard(Size);
                            end;
                        finally
                            ConnectedClient.DownloadList.UnLock;
                        end;
                        //Exit;
                    end;

                    //* Client sent a command
                    if SameText(Command[1], '700')
                    then begin
                        MessageCommand := TMessageCommand.Create;
                        MessageCommand.ID := StrToIntDef(Command[0], 0);
                        MessageCommand.Command := Command[2];
                        MessageCommand.Arguments := Copy(ClientData, Pos(Command[2], ClientData) + Length(Command[2]) + 1, Length(ClientData));
                        MessageCommand.Client := ConnectedClient;
                        MessageCommand.Context := AContext;
                        IncomingMessageQueue.Add(MessageCommand);
                        Exit;
                    end;

                end;

            end;

            if Assigned(ConnectedClient) then begin
                //* Is there binary to initialize/send?
                ConnectedClient.SendList.Lock;
                try
                    i := - 1;
                    while i <= ConnectedClient.SendList.Count - 2 do begin
                        Inc(i);
                        if TObject(ConnectedClient.SendList.List[i]) is TMessageBinary then begin
                            DoBinaryTransfer(AContext, TMessageBinary(ConnectedClient.SendList.List[i]), ConnectedClient.SendList);
                            Continue;
                        end;

                        //* Send chat text
                        if TObject(ConnectedClient.SendList.List[i]) is TMessageChat then begin
                            MessageChat := TMessageChat(ConnectedClient.SendList.List[i]);
                            AContext.Connection.IOHandler.WriteLn('0 111 ' + MessageChat.UserName + ': ' + MessageChat.Text, ByteEncoding);
                            ConnectedClient.SendList.List.Delete(i);
                            Dec(i);
                            {$IFNDEF NEXTGEN}
                            FreeAndNil(MessageChat);
                            {$ENDIF}
                            Continue;
                        end;

                        //* Send command
                        if TObject(ConnectedClient.SendList.List[i]) is TMessageCommand then begin
                            MessageCommand := TMessageCommand(ConnectedClient.SendList.List[i]);
                            AContext.Connection.IOHandler.WriteLn(IntToStr(MessageCommand.ID) + ' 700 ' + MessageCommand.Command + ' ' + MessageCommand.Arguments, ByteEncoding);
                            ConnectedClient.SendList.List.Delete(i);
                            Dec(i);
                            {$IFNDEF NEXTGEN}
                            FreeAndNil(MessageCommand);
                            {$ENDIF}
                            Continue;
                        end;
                    end;
                finally
                    ConnectedClient.SendList.UnLock;
                end;

                //* Is there binary to initialize/send synchronous?
                ConnectedClient.SendListSynchronous.Lock;
                try
                    if ConnectedClient.SendListSynchronous.Count > 0 then begin
                        if TObject(ConnectedClient.SendListSynchronous.List[0]) is TMessageBinary then begin
                            DoBinaryTransfer(AContext, TMessageBinary(ConnectedClient.SendListSynchronous.List[0]), ConnectedClient.SendListSynchronous);
                        end;
                    end;
                finally
                    ConnectedClient.SendListSynchronous.UnLock;
                end;

                //* Is there binary transfer to initialize/cancel?
                ConnectedClient.DownloadList.Lock;
                try
                    for i := ConnectedClient.DownloadList.Count - 1 downto 0 do begin
                        if TObject(ConnectedClient.DownloadList.List[i]) is TMessageBinary then begin
                            if NOT TMessageBinary(ConnectedClient.DownloadList.List[i]).RequestSent then begin
                                AContext.Connection.IOHandler.WriteLn(IntToStr(TMessageBinary(ConnectedClient.DownloadList.List[i]).ID) + ' 200 801', ByteEncoding);
                                TMessageBinary(ConnectedClient.DownloadList.List[i]).RequestSent := True;
                            end;
                        end;
                        if TObject(ConnectedClient.DownloadList.List[i]) is TMessageTransferCancel then begin
                            MessageTransferCancel := TMessageTransferCancel(ConnectedClient.DownloadList.List[i]);
                            //* Send a cancel notification
                            AContext.Connection.IOHandler.WriteLn(IntToStr(MessageTransferCancel.ID) + ' 809', ByteEncoding);
                            MessageBinaryIndex := ConnectedClient.DownloadList.Find(MessageTransferCancel.ID);
                            if MessageBinaryIndex > - 1 then begin
                                {$IFNDEF NEXTGEN}
                                TMessageTransferCancel(ConnectedClient.DownloadList.List[MessageBinaryIndex]).Free;
                                {$ENDIF}
                                ConnectedClient.DownloadList.Delete(MessageBinaryIndex);
                            end;
                        end;
                    end;
                finally
                    ConnectedClient.DownloadList.UnLock;
                end;
            end;

            if Assigned(AContext) then begin
                //* Heartbeat send
                if Assigned(TConnectionInfo(AContext.{$IFDEF TCPSCL_GENERICS_CONTEXTLIST}DataObject{$ELSE}Data{$ENDIF})) then begin
                    if MilliSecondsBetween(Now, TConnectionInfo(AContext.{$IFDEF TCPSCL_GENERICS_CONTEXTLIST}DataObject{$ELSE}Data{$ENDIF}).LastHeartBeatDateTime ) > FParameters.ClientPingInterval then begin
                        AContext.Connection.IOHandler.WriteLn('PING', ByteEncoding);
                        TConnectionInfo(AContext.{$IFDEF TCPSCL_GENERICS_CONTEXTLIST}DataObject{$ELSE}Data{$ENDIF}).LastHeartBeatDateTime := Now;
                    end;
                end;
            end;

        finally
            {$IFNDEF NEXTGEN}
            FreeAndNil(Command);
            {$ENDIF}
        end;
    except
        On E: Exception do begin
            if ClientData = '' then begin
                Log('Exception: ' + E.Message, True);
            end else begin
                Log('Exception: ' + E.Message + ' -> ' + ClientData, True);
            end;
            AContext.Connection.Disconnect;
        end;
    end;
end;

{$IFDEF TCPSCL_DELPHI_XE5 OR TCPSCL_MOBILE}
procedure TTCPServerThread.SSLGetPassword(var Password: String);
begin
    Password := FSSLPassword;
end;
{$ELSE}
procedure TTCPServerThread.SSLGetPassword(var Password: AnsiString);
begin
    Password := AnsiString(FSSLPassword);
end;
{$ENDIF}

procedure TTCPServerThread.Log(const Text: String; Debug: Boolean);
var
    MessageLog: TMessageLog;
begin
    if NOT LoggingOn then begin
        Exit;
    end;
    MessageLog := TMessageLog.Create;
    MessageLog.Text := Text;
    MessageLog.Debug := Debug;
    IncomingMessageQueue.Add(MessageLog);
end;

function TTCPServerThread.Active: Boolean;
begin
    Result := IdTCPServer.Active;
end;

procedure TTCPServerThread.SendBinary(MessageBinary: TMessageBinary);
var
    i: Integer;
    {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
    ContextList: System.Generics.Collections.TList<TIdContext>;
    {$ELSE}
    ContextList: TList;
    {$ENDIF}
    Context: TIdContext;
    ConnectedClient: TConnectedClient;
begin
    try
        if MessageBinary.Data.Size > 0 then begin
            ContextList := IdTCPServer.Contexts.LockList;
            try
                for i := 0 to ContextList.Count - 1 do begin
                    Context := TObject(ContextList[i]) as TIdContext;
                    if (MessageBinary.Context = Context)
                    OR (MessageBinary.Context = nil)
                    then begin
                        ConnectedClients.Lock;
                        try
                            ConnectedClient := ConnectedClients.Find(Context);
                            if Assigned(ConnectedClient) then begin
                                case MessageBinary.TransferType of
                                    ttAsynchronous: begin
                                        ConnectedClient.SendList.Lock;
                                        ConnectedClient.SendListSynchronous.Lock;
                                        ConnectedClient.DownloadList.Lock;
                                        try
                                            if MessageBinary.ID = 0 then begin
                                                MessageBinary.ID := ConnectedClient.GenerateUniqueTransferID;
                                            end;
                                            if MessageBinary.RequireAgreement = raFalse then begin
                                                MessageBinary.RequestSent := True;
                                                MessageBinary.DoTransfer := True;
                                            end;
                                            ConnectedClient.SendList.Add(MessageBinary);
                                        finally
                                            ConnectedClient.SendList.UnLock;
                                            ConnectedClient.SendListSynchronous.UnLock;
                                            ConnectedClient.DownloadList.UnLock;
                                        end;
                                    end;
                                    ttSynchronous: begin
                                        ConnectedClient.SendList.Lock;
                                        ConnectedClient.SendListSynchronous.Lock;
                                        ConnectedClient.DownloadList.Lock;
                                        try
                                            if MessageBinary.ID = 0 then begin
                                                MessageBinary.ID := ConnectedClient.GenerateUniqueTransferID;
                                            end;
                                            if MessageBinary.RequireAgreement = raFalse then begin
                                                MessageBinary.RequestSent := True;
                                                MessageBinary.DoTransfer := True;
                                            end;
                                            ConnectedClient.SendListSynchronous.Add(MessageBinary);
                                        finally
                                            ConnectedClient.SendList.UnLock;
                                            ConnectedClient.SendListSynchronous.UnLock;
                                            ConnectedClient.DownloadList.UnLock;
                                        end;
                                    end;
                                end;
                            end;
                        finally
                            ConnectedClients.UnLock;
                        end;
                    end;
                end;
            finally
                IdTCPServer.Contexts.UnlockList;
            end;
        end;
    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(MessageBinary);
        {$ENDIF}
    end;
end;

procedure TTCPServerThread.SendCommand(MessageCommand: TMessageCommand);
var
    i: Integer;
    {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
    ContextList: System.Generics.Collections.TList<TIdContext>;
    {$ELSE}
    ContextList: TList;
    {$ENDIF}
    Context: TIdContext;
    ConnectedClient: TConnectedClient;
begin
    try
        if MessageCommand.Command <> '' then begin
            ContextList := IdTCPServer.Contexts.LockList;
            try
                for i := 0 to ContextList.Count - 1 do begin
                    Context := TObject(ContextList[i]) as TIdContext;
                    if (MessageCommand.Context = Context)
                    OR (MessageCommand.Context = nil)
                    then begin
                        ConnectedClients.Lock;
                        try
                            ConnectedClient := ConnectedClients.Find(Context);
                            if Assigned(ConnectedClient) then begin
                                //* Send the command
                                ConnectedClient.SendList.Lock;
                                try
                                    ConnectedClient.SendList.Add(MessageCommand);
                                finally
                                    ConnectedClient.SendList.UnLock;
                                end;
                                //* Store the command
                                ConnectedClient.CommandList.Lock;
                                try
                                    ConnectedClient.CommandList.Add(MessageCommand);
                                finally
                                    ConnectedClient.CommandList.UnLock;
                                end;
                            end;
                        finally
                            ConnectedClients.UnLock;
                        end;
                    end;
                end;
            finally
                IdTCPServer.Contexts.UnlockList;
            end;
        end;
    finally
        //FreeAndNil(MessageCommand); //* Is stored in CommandList
    end;
end;

procedure TTCPServerThread.SendChat(MessageChat: TMessageChat);
var
    i: Integer;
    {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
    ContextList: System.Generics.Collections.TList<TIdContext>;
    {$ELSE}
    ContextList: TList;
    {$ENDIF}
    Context: TIdContext;
    ConnectedClient: TConnectedClient;
begin
    if MessageChat.Text <> '' then begin
        ContextList := IdTCPServer.Contexts.LockList;
        try
            for i := 0 to ContextList.Count - 1 do begin
                Context := TObject(ContextList[i]) as TIdContext;
                ConnectedClients.Lock;
                try
                    ConnectedClient := ConnectedClients.Find(Context);
                    if Assigned(ConnectedClient) then begin
                        ConnectedClient.SendList.Lock;
                        try
                            if MessageChat.UserName = '' then begin
                                MessageChat.UserName := FServerName;
                            end;
                            ConnectedClient.SendList.Add(MessageChat);
                        finally
                            ConnectedClient.SendList.UnLock;
                        end;
                    end;
                finally
                    ConnectedClients.UnLock;
                end;
            end;
        finally
            IdTCPServer.Contexts.UnlockList;
        end;
    end;
end;

procedure TTCPServerThread.SendClientList;
var
    i: Integer;
    NewClientList: TStringList;
    NewClientListMessageBinaryMemory: TMessageBinaryMemory;
begin
    ConnectedClients.Lock;
    try
        NewClientList := TStringList.Create;
        NewClientListMessageBinaryMemory := TMessageBinaryMemory.Create;
        try
            ConnectedClients.GenerateClientList(NewClientList);
            NewClientListMessageBinaryMemory.FileName := 'Client list';
            NewClientListMessageBinaryMemory.DataType := TCP_SOCKET_DATA_TYPE_CLIENT_LIST;
            NewClientListMessageBinaryMemory.RequireAgreement := raFalse;
            NewClientListMessageBinaryMemory.DoTransfer := True;
            NewClientList.SaveToStream(NewClientListMessageBinaryMemory.Data);
            NewClientListMessageBinaryMemory.TotalSize := NewClientListMessageBinaryMemory.Data.Size;
            for i := 0 to ConnectedClients.List.Count - 1 do begin
                if ConnectedClients.List[i].ExpectsClientList then begin
                    NewClientListMessageBinaryMemory.ID := ConnectedClients.List[i].GenerateUniqueTransferID;
                    ConnectedClients.List[i].SendListSynchronous.Add(NewClientListMessageBinaryMemory);
                end;
            end;
        finally
            FreeAndNil(NewClientListMessageBinaryMemory);
            FreeAndNil(NewClientList);
        end;
    finally
        ConnectedClients.UnLock;
    end;
end;

procedure TTCPServerThread.LoopbackChat(MessageChat: TMessageChat);
var
    ChatItem: TMessageChat;
begin
    with MessageChat as TMessageChat do begin
        ChatItem := TMessageChat.Create;
        ChatItem.UserName := UserName;
        ChatItem.Text := Text;
    end;
    IncomingMessageQueue.Add(ChatItem);
end;

procedure TTCPServerThread.SendTerminated;
var
    MessageTerminated: TMessageTerminated;
begin
    MessageTerminated := TMessageTerminated.Create;
    MessageTerminated.Terminated := True;
    IncomingMessageQueue.Add(MessageTerminated);
end;

procedure TTCPServerThread.ReportProgress(Context: TIdContext; Progress: Integer; TransferID: Integer; ElapsedTime, EstimatedTime: TDateTime; TransferRate: Cardinal);
var
    MessageProgress: TMessageProgress;
begin
    MessageProgress := TMessageProgress.Create;
    MessageProgress.ID := TransferID;
    MessageProgress.Context := Context;
    MessageProgress.Progress := Progress;
    MessageProgress.ElapsedTime := ElapsedTime;
    MessageProgress.EstimatedTime := EstimatedTime;
    if EstimatedTime = 0 then begin
        MessageProgress.RemainingTime := 0;
    end else begin
        MessageProgress.RemainingTime := EstimatedTime - ElapsedTime;
    end;
    MessageProgress.TransferRate := TransferRate;
    IncomingMessageQueue.Add(MessageProgress);
end;

procedure TTCPServerThread.ProcessTransferRequest(MessageTransferRequest: TMessageTransferRequest);
var
    MessageBinary: TMessageBinary;
    ConnectedClient: TConnectedClient;
begin
    try
        if MessageTransferRequest.Accept then begin
            IdTCPServer.Contexts.LockList;
            try
                ConnectedClients.Lock;
                try
                    ConnectedClient := ConnectedClients.Find(MessageTransferRequest.Context);
                    if Assigned(ConnectedClient) then begin
                        if MessageTransferRequest.SaveToFile then begin
                            MessageBinary := TMessageBinaryFile.Create(MessageTransferRequest.SaveToFileName, MessageTransferRequest.Resume);
                        end else begin
                            MessageBinary := TMessageBinaryMemory.Create;
                        end;
                        MessageBinary.DataType := MessageTransferRequest.DataType;
                        MessageBinary.ID := MessageTransferRequest.ID;
                        MessageBinary.TotalSize := MessageTransferRequest.TotalSize;
                        MessageBinary.StartPosition := MessageTransferRequest.StartPosition;
                        MessageBinary.FileName := MessageTransferRequest.FileName;
                        MessageBinary.User := MessageTransferRequest.User;
                        ConnectedClient.DownloadList.Lock;
                        ConnectedClient.DownloadList.Add(MessageBinary);
                        ConnectedClient.DownloadList.UnLock;
                    end;
                finally
                    ConnectedClients.UnLock;
                end;
            finally
                IdTCPServer.Contexts.UnlockList;
            end;
        end;
    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(MessageTransferRequest);
        {$ENDIF}
    end;
end;

procedure TTCPServerThread.TransferCancel(MessageTransferCancel: TMessageTransferCancel);
var
    {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
    ContextList: System.Generics.Collections.TList<TIdContext>;
    {$ELSE}
    ContextList: TList;
    {$ENDIF}
    Context: TIdContext;
    ConnectedClient: TConnectedClient;
    i: Integer;
    Index: Integer;
begin
    try
        if MessageTransferCancel.ID <> 0 then begin
            ContextList := IdTCPServer.Contexts.LockList;
            try
                for i := 0 to ContextList.Count - 1 do begin
                    Context := TObject(ContextList[i]) as TIdContext;
                    if (MessageTransferCancel.Context = Context)
                    //OR (MessageTransferCancel.Context = nil)  // should not happen as every context uses a different ID
                    then begin
                        ConnectedClients.Lock;
                        try
                            ConnectedClient := ConnectedClients.Find(Context);
                            if Assigned(ConnectedClient) then begin
                                //* Send the cancel
                                ConnectedClient.DownloadList.Lock;
                                try
                                    ConnectedClient.DownloadList.Add(TMessageBinary(MessageTransferCancel));
                                finally
                                    ConnectedClient.DownloadList.UnLock;
                                end;
                                //* Local transfer remove
                                ConnectedClient.SendList.Lock;
                                try
                                    Index := ConnectedClient.SendList.Find(MessageTransferCancel.ID);
                                    if Index > - 1 then begin
                                        {$IFNDEF NEXTGEN}
                                        TObject(ConnectedClient.SendList.List[Index]).Free;
                                        {$ENDIF}
                                        ConnectedClient.SendList.Delete(Index);
                                    end;
                                finally
                                    ConnectedClient.SendList.UnLock;
                                end;
                                ConnectedClient.SendListSynchronous.Lock;
                                try
                                    Index := ConnectedClient.SendListSynchronous.Find(MessageTransferCancel.ID);
                                    if Index > - 1 then begin
                                        {$IFNDEF NEXTGEN}
                                        TObject(ConnectedClient.SendListSynchronous.List[Index]).Free;
                                        {$ENDIF}
                                        ConnectedClient.SendListSynchronous.Delete(Index);
                                    end;
                                finally
                                    ConnectedClient.SendListSynchronous.UnLock;
                                end;
                            end;
                        finally
                            ConnectedClients.UnLock;
                        end;
                    end;
                end;
            finally
                IdTCPServer.Contexts.UnlockList;
            end;
        end;
    finally
        //*
    end;
end;

procedure TTCPServerThread.ClientKick(MessageClientKick: TMessageClientKick);
var
    {$IFDEF TCPSCL_GENERICS_CONTEXTLIST}
    ContextList: System.Generics.Collections.TList<TIdContext>;
    {$ELSE}
    ContextList: TList;
    {$ENDIF}
    Context: TIdContext;
    i: Integer;
begin
    try
        if MessageClientKick.Context <> nil then begin
            ContextList := IdTCPServer.Contexts.LockList;
            try
                for i := 0 to ContextList.Count - 1 do begin
                    Context := TObject(ContextList[i]) as TIdContext;
                    if (MessageClientKick.Context = Context) then begin
                        Context.Connection.Disconnect;
                        Break;
                    end;
                end;
            finally
                IdTCPServer.Contexts.UnlockList;
            end;
        end;
    finally
        {$IFNDEF NEXTGEN}
        FreeAndNil(MessageClientKick);
        {$ENDIF}
    end;
end;

end.

